# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-03-18 14:42:17
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2022-08-15 09:28:52

import numpy as np


def bounds(x):
    '''Get the bounds of an array ''' 
    return np.array([x.min(), x.max()])


def nan_like(x):
    ''' Create array of identicial shape as input, filled with nan values '''
    xnan = np.zeros_like(x)
    xnan[:] = np.nan
    return xnan


def is_within(x, bounds):
    return np.logical_and(x >= bounds[0], x <= bounds[1])



# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2021-10-11 13:30:15
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2022-03-16 13:02:25

''' Collection of logging utilities. '''

import sys
import colorlog
import logging
import tqdm

my_log_formatter = colorlog.ColoredFormatter(
    '%(log_color)s %(asctime)s %(message)s',
    datefmt='%d/%m/%Y %H:%M:%S:',
    reset=True,
    log_colors={
        'DEBUG': 'green',
        'INFO': 'white',
        'WARNING': 'yellow',
        'ERROR': 'red',
        'CRITICAL': 'red,bg_white',
    },
    style='%')


def setHandler(logger, handler):
    for h in logger.handlers:
        logger.removeHandler(h)
    logger.addHandler(handler)
    return logger


def setLogger(name, formatter):
    handler = colorlog.StreamHandler()
    handler.setFormatter(formatter)
    handler.stream = sys.stdout
    logger = colorlog.getLogger(name)
    logger.addHandler(handler)
    return logger


class TqdmHandler(logging.StreamHandler):

    def __init__(self, formatter):
        logging.StreamHandler.__init__(self)
        self.setFormatter(formatter)

    def emit(self, record):
        msg = self.format(record)
        tqdm.write(msg)


logger = setLogger('mylogger', my_log_formatter)
logger.setLevel(logging.INFO)



from nidaqmx.system import System
from nidaqmx.system.device import Device
from nidaqmx.task import Task
import numpy as np

from .logger import logger


def get_nidaq_device_names():
    '''  Get NI DAQ device names list '''
    return System().devices.device_names

def get_nidaq_terminals(device_name):
    ''' Get NI DAQ device output terminals list '''
    return Device(device_name).terminals


def get_nidaq_tasks():
    return System().tasks.task_names


class PulseTrainGenerator:
    ''' 
    Interface to generate a train of digital output pulses on the NI DAQ with
    specified frequency and initial delay relative to each supplied input triggers
    '''

    timeout = 5

    ######################### CONSTRUCTORS & DESTRUCTORS #########################

    def __init__(self, device_name=None, input_PFI=0, output_PFI=1, task_name=None, 
                 freq=1., duty_cycle=0.5, initial_delay=0, npulses=1):
        '''
        Constructor

        :param device_name: DAQmx device name (e.g. "PXI1Slot6")
        :param input_PFI: PFI slot number on which input trigger is received
        :param output_PFI: PFI slot number on which output pulse(s) is generated
        :param freq: frequency (in Hz) at which to generate pulses.
        :param duty_cycle: width of pulses divided by the inter-pulse period.
        :param initial_delay: time in seconds to wait before generating the first pulse.
        :param npulses: number of pulses to generate
        '''
        # DAQ parameters 
        self.device_name = device_name
        self.input_PFI = input_PFI
        self.output_PFI = output_PFI
        self.task_name = task_name
        # Pulse train parameters
        self.freq = freq
        self.duty_cycle = duty_cycle
        self.initial_delay = initial_delay
        self.npulses = npulses
        # Create task
        self.create_task()
        # Log upon creation
        logger.info(f'created {self}')
    
    def __del__(self):
        ''' Destructor '''
        # Close underlying task
        if hasattr(self, '_task'):
            logger.info(f'closing {self.task_name} task ...')
            self._task.close()
            del(self._task)
    
    def __repr__(self):
        return f'{self.__class__.__name__}({self.device_name}, PFI_in={self.input_PFI}, PFI_out={self.output_PFI}, freq={self.freq:.2f}Hz, delay={self.initial_delay:.2f}s, npulses={self.npulses})'

    ######################### GETTERS & SETTERS #########################

    @staticmethod
    def validate_attribute(key, val, dtype=float):
        if not isinstance(val, dtype):
            if dtype == float and isinstance(val, int):
                val = float(val)
            else:
                raise ValueError(f'{key} is not a {dtype}')
        if val < 0 or val == np.inf:
            raise ValueError(f'{key} must be a positive finite scalar')

    @property
    def device_name(self):
        return self._device_name
        
    @device_name.setter
    def device_name(self, val):
        if val is None:
            val = get_nidaq_device_names()[0]
        self._device_name = val
            
    @property
    def input_PFI(self):
        return self._input_PFI
    
    @input_PFI.setter
    def input_PFI(self, val):
        if val is not None:
            self.validate_attribute('input_PFI', val, dtype=int)
        self._input_PFI = val
        self.__update_trigger_props('input_PFI')
    
    @property
    def output_PFI(self):
        return self._output_PFI
    
    @output_PFI.setter
    def output_PFI(self, val):
        if val is not None:
            self.validate_attribute('output_PFI', val, dtype=int)
        self._output_PFI = val
        self.__update_trigger_props('output_PFI')
    
    @property
    def task_name(self):
        return self._task_name
    
    @task_name.setter
    def task_name(self, val):
        if val is None:
            val = self.__class__.__name__
        self._task_name = val
    
    @property
    def freq(self):
        return self._freq
    
    @freq.setter
    def freq(self, val):
        self.validate_attribute('freq', val)
        self._freq = val
        self.__update_timer_props('freq')
    
    @property
    def duty_cycle(self):
        return self._duty_cycle
    
    @duty_cycle.setter
    def duty_cycle(self, val):
        self.validate_attribute('duty_cycle', val)
        self._duty_cycle = val
        self.__update_timer_props('duty_cycle')
    
    @property
    def initial_delay(self):
        return self._initial_delay
    
    @initial_delay.setter
    def initial_delay(self, val):
        self.validate_attribute('initial_delay', val)
        self._initial_delay = val
        self.__update_timer_props('initial_delay')
    
    @property
    def npulses(self):
        return self._npulses
    
    @npulses.setter
    def npulses(self, val):
        self.validate_attribute('npulses', val, dtype=int)
        self._npulses = val
        self.__update_timer_props('npulses')

    ######################### DERIVED PROPERTIES #########################

    @property   
    def input_terminal(self):
        if self.input_PFI is None:
            return get_nidaq_terminals(self.device_name)[0]
        return f'/{self.device_name}/PFI{self.input_PFI}'

    @property   
    def output_terminal(self):
        if self.output_PFI is None:
            return get_nidaq_terminals(self.device_name)[1]
        return f'/{self.device_name}/PFI{self.output_PFI}'

    @property
    def output_channel(self):
        return f'{self.device_name}/ctr{self.output_PFI}'
    
    ######################### OTHER METHODS #########################

    def create_task(self):
        ''' Create underlying task '''
        self._task = Task(self.task_name)
        self._task.co_channels.add_co_pulse_chan_freq(
            self.output_channel,
            initial_delay=self.initial_delay,
            freq=self.freq,
            duty_cycle=self.duty_cycle,
        )
        self.__update_timer_props('init')
        self.__update_trigger_props('init')
        
    def check_task_disabled(self, propName):
        ''' Throw error if task is already enabled while trying to set a property '''
        assert self._task.is_task_done(), f'Cannot update property "{propName}" while pulse train generator is enabled'

    def __update_timer_props(self, propName):
        ''' Propagate timing properties to underlying task & channel '''
        # If underlying task exists
        if hasattr(self, '_task'):
            # Check that task is disabled
            self.check_task_disabled(propName)
            # Set Task channel pulse properties: frequency, duty cycle,
            # delay, and number of pulses
            self._task.co_channels[0].co_pulse_freq = self.freq
            self._task.co_channels[0].co_pulse_duty_cyc = self.duty_cycle
            self._task.co_channels[0].co_pulse_freq_initial_delay = self.initial_delay
            self._task.timing.cfg_implicit_timing(samps_per_chan= self.npulses)
            # Ensure that initial delay is conserved upon re-triggering
            self._task.co_channels[0].co_enable_initial_delay_on_retrigger = True
        
    def __update_trigger_props(self, propName):
        ''' Propagate trigger properties to underlying task '''
        # If underlying task exists
        if hasattr(self, '_task'):
            # Check that task is disabled
            self.check_task_disabled(propName)
            # Set the task to start upon acquisition start
            self._task.triggers.start_trigger.cfg_dig_edge_start_trig(self.input_terminal)
            # Set task to be triggerable
            self._task.triggers.start_trigger.retriggerable = True
            # Set pulse output terminal
            self._task.co_channels[0].co_pulse_term = self.output_terminal
    
    def enable(self):
        assert self._task.is_task_done(), 'Pulse train generator is already enabled'
        logger.info(f'starting {self.task_name} task...')
        self._task.start()
        
    def disable(self):
        logger.info(f'stopping {self.task_name} task...')
        self._task.stop()


def get_trigger_pulses(name, delay=0, interval=1., npulses=1, PFI=1):
    '''
    Set up a train of TTL pulse(s) to be triggered upon acquisition 
    
    :param name: unique string identifier of the underlying NI-DAQ task
    :param delay: delay in seconds to initiate first pulse (default = 0)
    :param interval: interval in seconds between pulses (default = 1s)
    :param npulses: number of pulses (default = 1)
    :param PFI: output PFI slot number (default = 1)
    :return: PulseTrainGenerator object
    '''    
    # Create PulseTrainGenerator object
    ptg = PulseTrainGenerator(
        device_name='PXI1Slot6', # Device name, 
        input_PFI=0, output_PFI=PFI, # Input & output PFI slot numbers
        task_name=name,
        freq=1 / interval, # frequency set to reciprocal of inter-pulse interval
        duty_cycle= .005 / interval,  # duty cycle set to obtain a constant pusle width of 5 ms
        initial_delay=delay,
        npulses=npulses
    )
    # Start underlying task
    ptg.enable()
    # Enable
    return ptg


# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-08-15 10:21:10
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-17 10:54:10

''' List all available VISA resources '''

from instrulink import list_visa_resources

list_visa_resources()



# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-11-02 12:00:31
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2022-11-02 12:22:59

from .visa_instrument import *

class KeysightE5061B(VisaInstrument):
    ''' Interface Keysight E5061B ENA Network Analyzer using the SCPI command interface '''

    PREFIX = ':'
    USB_ID = 'MY49810002'
    NO_ERROR_CODE = '0,"No error"'  # not sure

    def reset(self):
        pass

    def clear(self):
        pass

    def get_last_error(self):
        pass
    
    def get_trigger_slope(self):
        pass
    
    def lock_front_panel(self):
        pass
    
    def set_trigger_slope(self):
        pass
    
    def unlock_front_panel(self):
        pass
    
    def get_channel_data(self, ich=1):
        '''
        Extract data from a given channel
        
        :param ich: channel index
        :return: 2-tuple with:
            - frequency vector (Hz)
            - values vector (???)
        '''
        self.write('FORM:DATA ASC')
        m = self.ask_for_values(f'CALC{ich}:DATA:FDAT?')
        m = m[::2]
        f = self.ask_for_values(f'SENS{ich}:FREQ:DATA?')
        return f, m




# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2023-05-10 23:09:28
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-17 15:13:14

from .rigol_dg1022z import RigolDG1022Z
from .rigol_ds1054z import RigolDS1054Z
from .bk_2555 import BK2555
from .sutter_mp285a import SutterMP285A, SutterError
from .visa_instrument import VisaError
from .logger import logger

''' High-level interface functions to access lab instruments. '''

# Dictionary of available waveform generator classes
generator_classes = {
    'rigol': RigolDG1022Z
}

# Dictionary of available oscilloscope classes
oscilloscope_classes = {
    'bk': BK2555,
    'rigol': RigolDS1054Z
}

# Dictionary of available micro-manipulator classes
manipulator_classes = {
    'sutter': SutterMP285A
}


def grab_instrument(type, instdict, key=None):
    ''' 
    Generic function to grab instrument object.
    
    :param type: intrument type (e.g. generator, oscilloscope, etc.)
    :param instdict: dictionary of instrument classes for that specific type
    :param key: instrument key (optional, if None, the first available instrument is returned)
    :return: instrument object
    '''
    if key is not None:
        if key not in instdict.keys():
            raise ValueError(f'Invalid {type} key: "{key}". Candidates are: {instdict.keys()}')
        return instdict[key]()
    else:
        for key in instdict.keys():
            try:
                logger.info(f'Attempting to grab "{key}" {type} ...')
                return instdict[key]()
            except (VisaError, SutterError):
                logger.info(f'Failed to grab "{key}" {type}')
        raise ValueError(f'No {type} found')


def grab_generator(**kwargs):
    ''' Grab waveform generator '''
    return grab_instrument('waveform generator', generator_classes, **kwargs)


def grab_oscilloscope(**kwargs):
    ''' Grab oscilloscope '''
    return grab_instrument('oscilloscope', oscilloscope_classes, **kwargs)
    

def grab_manipulator(**kwargs):
    ''' Grab micro-manipulator '''
    return grab_instrument('micro-manipulator', manipulator_classes, **kwargs)



# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2023-04-25 13:25:00
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-17 10:54:12

from instrulink.nidaq import *

# Parameters
interval = 1.5 #20.019  # s
npulses = 16
camdelay = 0.  # s
stimdelay = 2.8  # s

# Set up train of trigger pulses for US stimuli
stim_triggers = get_trigger_pulses('stim_triggers', stimdelay , interval, npulses, 1)
print(stim_triggers)

# Set up train of trigger pulses for camera acquisitions
cam_triggers = get_trigger_pulses('cam_triggers', camdelay, interval, npulses, 2)
print(cam_triggers)



# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-03-08 08:37:26
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-08-08 15:17:25

import time
import re
from tqdm import tqdm
import numpy as np

from .waveform_generator import *
from .logger import logger
from .si_utils import si_format
from .constants import TTL_PWIDTH, TTL_PAMP, MV_TO_V
from .utils import is_within
from instrulink.wf_utils import *


class RigolDG1022Z(WaveformGenerator):

    # Miscellaneous
    USB_ID = 'DG1ZA\d+'  # USB identifier
    NO_ERROR_CODE = '0,"No error"'  # error code returned when no error
    ANGLE_UNITS = ('DEG')
    MAX_LEN_TEXT = 40
    WAVEFORM_TYPES = ('SIN', 'SQU', 'RAMP', 'PULS', 'NOIS', 'DC', 'USER')
    FMAX = 25e6  # max frequency (Hz)
    VMAX = 20.0  # max voltage (Vpp)
    ANTIPHASE = 180  # degrees
    CHANNELS = (1, 2)
    PREFIX = ':'
    # TIMEOUT_SECONDS = 20.  # long timeout to allow slow commands (e.g. waveform loading)

    # Coupling
    CPL_PATTERN = '^FREQ:(ON|OFF),PHASE:(ON|OFF),AMPL:(ON|OFF)$'
    CPL_MODES = ('OFFS', 'RAT')  # coupling modes
    CPL_AMP_RATIO_BOUNDS = (1e-3, 1e3)  # bounds for amplitude coupling ratio
    CPL_AMP_DEV_BOUNDS = (-19.998, 19.998)  # bounds for amplitude coupling deviation
    CPL_FREQ_RATIO_BOUNDS = (1e-6, 1e6)  # bounds for frequency coupling ratio
    CPL_FREQ_DEV_BOUNDS = (-0.99 * FMAX, 0.99 * FMAX)  # bounds for frequency coupling deviation
    CPL_PHASE_RATIO_BOUNDS = (1e-2, 1e2)  # bounds for phase coupling ratio
    CPL_PHASE_DEV_BOUNDS = (-360, 360)  # bounds for phase coupling deviation

    # Bursting
    BURST_MODES = ('TRIG', 'INF', 'GAT')
    BURST_IDLE_LEVELS = ('FPT', 'TOP', 'CENTER', 'BOTTOM')
    PULSE_HOLDS = ('WIDT', 'DUTY')
    MAX_BURST_PERIOD = 500. # s

    # Modulation
    MOD_MODES = ('AM', 'FM', 'PM', 'ASK', 'FSK', 'PSK', 'PWM')
    MOD_VOLT_RANGE = (-5, 5)  # voltage range regulating modulation (V)
    MOD_VOLT_MARGIN = 0.05  # margin on each side to ensure full range amplitude modulation (V)
    AM_DEPTH_RANGE = (0, 120)  # AM depth range (%)
    AM_FREQ_RANGE = (2e-3, 1e6)  # AM frequency range (Hz)
    MOD_SOURCES = ('INT', 'EXT')  # modulation sources

    # Trigger
    TRIGGER_SOURCES = ('INT', 'EXT', 'MAN')

    # Arbitrary waveform
    ARB_OUTPUT_MODES = ('FREQ', 'SRATE')
    ARB_SRATE_BOUNDS = (1e-6, 60e6)  # bounds for arbitrary waveform sample rate (Hz)
    ARB_WF_NPTS_BOUNDS = (8, 16384)  # bounds for arbitrary waveform number of points
    ARB_WF_MAXNPTS_PER_PACKET = 8192  # max number of points per packet for arbitrary waveform upload
    ARB_WF_DAC_RANGE = (0, 16383)  # integer range for arbitrary waveform DAC values
    ARB_WF_FLOAT_RANGE = (-1, 1)  # floating point range for arbitrary waveform values
    WAF_PATTERN = '^(ARB)(10?|[2-9])$'

    def beep(self):
        ''' Issue a single beep immediately. '''
        self.write('SYST:BEEP:IMM')
    
    # --------------------- UNITS ---------------------
    
    def set_angle_unit(self, unit):
        self.check_angle_unit(unit)

    def get_angle_unit(self):
        return self.ANGLE_UNITS[0]
    
    def set_voltage_unit(self, ich, unit):
        self.check_channel_index(ich)
        self.check_voltage_unit(unit)
        self.write(f'SOUR{ich}:VOLT:UNIT {unit}')

    def get_voltage_unit(self, ich):
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:VOLT:UNIT?')

    # --------------------- OUTPUT ---------------------

    def enable_output_channel(self, ich):
        ''' Turn channel output ON '''
        self.check_channel_index(ich)
        self.write(f'OUTP{ich} ON')

    def disable_output_channel(self, ich):
        ''' Turn channel output OFF '''
        self.check_channel_index(ich)
        self.write(f'OUTP{ich} OFF')   

    def enable_output(self):
        for ich in self.CHANNELS:
            self.enable_output_channel(ich)

    def disable_output(self):
        for ich in self.CHANNELS:
            self.disable_output_channel(ich)
    
    def get_output_state(self, *ich):
        ''' Get the output state for a specific channel '''
        if len(ich) == 0:
            ich = self.CHANNELS
        if len(ich) > 1:
            return [self.get_output_state(x) for x in ich]
        ich = ich[0]
        self.check_channel_index(ich)
        return self.query(f'OUTP{ich}?')

    def test_output(self):
        ''' Test output enabling/disabling sequence '''
        self.enable_output_channel(1)
        time.sleep(1)
        self.enable_output_channel(2)
        time.sleep(1)
        self.disable_output_channel(2)
        time.sleep(1)
        self.disable_output_channel(1)
        time.sleep(1)
        self.enable_output()
        time.sleep(1)
        self.disable_output()
    
    # --------------------- SYNC ---------------------
    
    def enable_output_sync(self, ich):
        ''' Enable sync signal of specific channel on rear panel connector '''
        self.check_channel_index(ich)
        self.write(f'OUTP{ich}:SYNC ON')
    
    def disable_output_sync(self, ich):
        ''' Disable sync signal of specific channel on rear panel connector '''
        self.check_channel_index(ich)
        self.write(f'OUTP{ich}:SYNC OFF')
    
    def is_output_sync_on(self, ich):
        ''' Query whether sync signal is enabled for a specific channel '''
        self.check_channel_index(ich)
        out = self.query(f'OUTP{ich}:SYNC?')
        return out == 'ON'

    def set_output_sync_delay(self, ich, delay):
        ''' 
        Set sync signal delay for a specific channel 

        :param ich: channel index
        :param delay: delay (s)
        '''
        self.check_channel_index(ich)
        T = 1 / self.get_waveform_freq(ich)
        if not is_within(delay, (0, T)):
            raise VisaError(f'invalid delay: {delay} (must be within [0, 1/f = {si_format(T, 2)}s])')
        self.write(f'OUTP{ich}:SYNC:DEL {delay}')
    
    def get_output_sync_delay(self, ich):
        ''' 
        Get sync signal delay for a specific channel 

        :param ich: channel index
        :return: delay (s)
        '''
        self.check_channel_index(ich)
        return float(self.query(f'OUTP{ich}:SYNC:DEL?'))
    
    def set_output_sync_polarity(self, ich, pol):
        ''' 
        Set sync signal polarity for a specific channel 

        :param ich: channel index
        :param pol: polarity ("POS" or "NEG")
        '''
        self.check_channel_index(ich)
        if pol not in self.SLOPES:
            raise VisaError(f'invalid polarity: "{pol}" (must be in {self.SLOPES})')
        self.write(f'OUTP{ich}:SYNC:POL {pol}')

    def get_output_sync_polarity(self, ich):
        ''' 
        Get sync signal polarity for a specific channel 

        :param ich: channel index
        :return: polarity ("POS" or "NEG")
        '''
        self.check_channel_index(ich)
        return self.query(f'OUTP{ich}:SYNC:POL?')

    # --------------------- COUPLING ---------------------

    def enable_all_coupling(self):
        ''' 
        Enable frequency, phase and amplitude coupling across channels.
        '''
        self.write('COUP ON')
    
    def disable_all_coupling(self):
        ''' 
        Disable frequency, phase and amplitude coupling across channels.
        '''
        self.write('COUP OFF')
    
    def is_coupling_on(self):
        ''' 
        Query whether frequency, phase and amplitude coupling across channels are enabled.

        :return: dictionary of coupling states
        '''
        out = self.query('COUP?')
        mo = re.match(self.CPL_PATTERN, out)
        if mo is None:
            raise VisaError(f'invalid coupling query response: "{out}"')
        cplstates = dict(zip(('FREQ', 'PHASE', 'AMPL'), mo.groups()))
        return {k: v == 'ON' for k, v in cplstates.items()}

    # --------------------- AMPLITUDE COUPLING ---------------------

    def enable_amplitude_coupling(self):
        ''' 
        Enable amplitude coupling across channels.
        '''
        self.write('COUP:AMPL ON')
    
    def disable_amplitude_coupling(self):
        ''' 
        Disable amplitude coupling across channels.
        '''
        self.write('COUP:AMPL OFF')
    
    def is_amplitude_coupling_on(self):
        ''' 
        Query whether amplitude coupling across channels is enabled.

        :return: boolean
        '''
        return self.query('COUP:AMPL?') == 'ON'

    def set_amplitude_coupling_mode(self, mode):
        ''' 
        Set amplitude coupling mode across channels.

        :param mode: amplitude coupling ("OFFS" for deviation or "RAT" for ratio)
        '''
        if mode not in self.CPL_MODES:
            raise VisaError(f'invalid amplitude coupling mode: "{mode}" (must be one of {self.CPL_MODES})')
        self.write(f'COUP:AMPL:MODE {mode}')
    
    def get_amplitude_coupling_mode(self):
        ''' 
        Get amplitude coupling mode across channels.

        :return: amplitude coupling mode
        '''
        return self.query('COUP:AMPL:MODE?')

    def set_amplitude_coupling_ratio(self, ratio):
        ''' 
        Set amplitude coupling ratio across channels.

        :param ratio: amplitude coupling ratio
        '''
        if not is_within(ratio, self.CPL_AMP_RATIO_BOUNDS):
            raise VisaError(f'invalid amplitude coupling ratio: "{ratio}" (must be within {self.CPL_AMP_RATIO_BOUNDS})')
        self.write(f'COUP:AMPL:RAT {ratio}')
    
    def get_amplitude_coupling_ratio(self):
        ''' 
        Get amplitude coupling ratio across channels.

        :return: amplitude coupling ratio
        '''
        return float(self.query('COUP:AMPL:RAT?'))

    def set_amplitude_coupling_deviation(self, deviation):
        ''' 
        Set amplitude coupling deviation across channels.

        :param deviation: amplitude coupling deviation
        '''
        if not is_within(deviation, self.CPL_AMP_DEV_BOUNDS):
            raise VisaError(f'invalid amplitude coupling deviation: "{deviation}" (must be within {self.CPL_AMP_DEV_BOUNDS})')
        self.write(f'COUP:AMPL:DEV {deviation}')
    
    def get_amplitude_coupling_deviation(self):
        ''' 
        Get amplitude coupling deviation across channels.

        :return: amplitude coupling deviation
        '''
        return float(self.query('COUP:AMPL:DEV?'))
    
    # --------------------- FREQUENCY COUPLING ---------------------

    def enable_frequency_coupling(self):
        ''' 
        Enable frequency coupling across channels.
        '''
        self.write('COUP:FREQ ON')

    def disable_frequency_coupling(self):
        ''' 
        Disable frequency coupling across channels.
        '''
        self.write('COUP:FREQ OFF')

    def is_frequency_coupling_on(self):
        '''
        Query whether frequency coupling across channels is enabled.
        '''
        return self.query('COUP:FREQ?') == 'ON'

    def set_frequency_coupling_mode(self, mode):
        ''' 
        Set frequency coupling mode across channels.

        :param mode: frequency coupling ("OFFS" for deviation or "RAT" for ratio)
        '''
        if mode not in self.CPL_MODES:
            raise VisaError(f'invalid frequency coupling mode: "{mode}" (must be one of {self.CPL_MODES})')
        self.write(f'COUP:FREQ:MODE {mode}')
    
    def get_frequency_coupling_mode(self):
        ''' 
        Get frequency coupling mode across channels.

        :return: frequency coupling mode
        '''
        return self.query('COUP:FREQ:MODE?')

    def set_frequency_coupling_ratio(self, ratio):
        ''' 
        Set frequency coupling ratio across channels.

        :param ratio: frequency coupling ratio
        '''
        if not is_within(ratio, self.CPL_FREQ_RATIO_BOUNDS):
            raise VisaError(f'invalid frequency coupling ratio: "{ratio}" (must be within {self.CPL_FREQ_RATIO_BOUNDS})')
        self.write(f'COUP:FREQ:RAT {ratio}')
    
    def get_frequency_coupling_ratio(self):
        ''' 
        Get frequency coupling ratio across channels.

        :return: frequency coupling ratio
        '''
        return float(self.query('COUP:FREQ:RAT?'))

    def set_frequency_coupling_deviation(self, deviation):
        ''' 
        Set frequency coupling deviation across channels.

        :param deviation: frequency coupling deviation
        '''
        if not is_within(deviation, self.CPL_FREQ_DEV_BOUNDS):
            raise VisaError(f'invalid frequency coupling deviation: "{deviation}" (must be within {self.CPL_FREQ_DEV_BOUNDS})')
        self.write(f'COUP:FREQ:DEV {deviation}')

    def get_frequency_coupling_deviation(self):
        ''' 
        Get frequency coupling deviation across channels.

        :return: frequency coupling deviation
        '''
        return float(self.query('COUP:FREQ:DEV?'))

    # --------------------- PHASE COUPLING ---------------------

    def enable_phase_coupling(self):
        ''' 
        Enable phase coupling across channels.
        '''
        self.write('COUP:PHAS ON')
    
    def disable_phase_coupling(self):
        ''' 
        Disable phase coupling across channels.
        '''
        self.write('COUP:PHAS OFF')
    
    def is_phase_coupling_on(self):
        '''
        Query whether phase coupling across channels is enabled.
        '''
        return self.query('COUP:PHAS?') == 'ON'

    def set_phase_coupling_mode(self, mode):
        ''' 
        Set phase coupling mode across channels.

        :param mode: phase coupling ("OFFS" for deviation or "RAT" for ratio)
        '''
        if mode not in self.CPL_MODES:
            raise VisaError(f'invalid phase coupling mode: "{mode}" (must be one of {self.CPL_MODES})')
        self.write(f'COUP:PHAS:MODE {mode}')
    
    def get_phase_coupling_mode(self):
        ''' 
        Get phase coupling mode across channels.

        :return: phase coupling mode
        '''
        return self.query('COUP:PHAS:MODE?')
    
    def set_phase_coupling_ratio(self, ratio):
        ''' 
        Set phase coupling ratio across channels.

        :param ratio: phase coupling ratio
        '''
        if not is_within(ratio, self.CPL_PHASE_RATIO_BOUNDS):
            raise VisaError(f'invalid phase coupling ratio: "{ratio}" (must be within {self.CPL_PHASE_RATIO_BOUNDS})')
        self.write(f'COUP:PHAS:RAT {ratio}')
    
    def get_phase_coupling_ratio(self):
        ''' 
        Get phase coupling ratio across channels.

        :return: phase coupling ratio
        '''
        return float(self.query('COUP:PHAS:RAT?'))
    
    def set_phase_coupling_deviation(self, deviation):
        ''' 
        Set phase coupling deviation across channels.

        :param deviation: phase coupling deviation
        '''
        if not is_within(deviation, self.CPL_PHASE_DEV_BOUNDS):
            raise VisaError(f'invalid phase coupling deviation: "{deviation}" (must be within {self.CPL_PHASE_DEV_BOUNDS})')
        self.write(f'COUP:PHAS:DEV {deviation}')
    
    def get_phase_coupling_deviation(self):
        ''' 
        Get phase coupling deviation across channels.

        :return: phase coupling deviation
        '''
        return float(self.query('COUP:PHAS:DEV?'))

    # --------------------- BASIC WAVEFORM ---------------------

    def apply_waveform(self, wtype, ich, freq, amp, offset=0, phase=0):
        ''' Apply a waveform with specific parameters.

            :param wtype: waveform function
            :param ich: channel number
            :param freq: waveform frequency (Hz)
            :param amp: waveform amplitude (default: Vpp)
            :param offset: voltage offset (V)
            :param phase: waveform phase (degrees)
        '''
        self.check_channel_index(ich)
        self.check_waveform_type(wtype)
        self.write(f'SOUR{ich}:APPL:{wtype} {freq}, {amp}, {offset}, {phase}')
    
    def apply_arbitrary(self, ich, sr):
        ''' 
        Apply an arbitrary waveform with specific sample rate. This function automatically
        sets the waveform type to "USER" and the output mode to "SRATE".

        :param ich: channel number
        :param sr: sample rate (samples/s)
        '''
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:APPL:ARB {sr}')

    def apply_noise(self, ich, amp, offset):
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:APPL:NOIS {amp}, {offset}')

    def set_waveform_type(self, ich, wtype):
        self.check_channel_index(ich)
        self.check_waveform_type(wtype)
        self.write(f'SOUR{ich}:FUNC {wtype}')

    def get_waveform_type(self, ich):
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:FUNC?')

    def set_waveform_freq(self, ich, freq):
        self.check_channel_index(ich)
        self.check_freq(freq)
        self.write(f'SOUR{ich}:FREQ:FIX {freq}')

    def get_waveform_freq(self, ich):
        self.check_channel_index(ich)
        return float(self.query(f'SOUR{ich}:FREQ:FIX?'))
    
    def set_waveform_amp(self, ich, amp):
        self.check_channel_index(ich)
        self.check_amp(amp)
        self.write(f'SOUR{ich}:VOLT:LEV:IMM:AMPL {amp}')

    def get_waveform_amp(self, ich):
        self.check_channel_index(ich)
        return float(self.query(f'SOUR{ich}:VOLT:LEV:IMM:AMPL?'))
    
    def check_offset(self, offset, ich):
        if np.absolute(offset) > self.VMAX - self.get_waveform_amp(ich) / 2:
            raise VisaError(f'|VPP/2| + |Voffset| exceeds {self.VMAX} V')

    def set_waveform_offset(self, ich, offset):
        self.check_channel_index(ich)
        self.check_offset(offset, ich)
        self.write(f'SOUR{ich}:VOLT:LEV:IMM:OFFS {offset}')

    def get_waveform_offset(self, ich):
        self.check_channel_index(ich)
        return float(self.query(f'SOUR{ich}:VOLT:LEV:IMM:OFFS?'))
    
    def set_waveform_phase(self, ich, phase):
        self.check_channel_index(ich)
        self.check_phase(phase)
        self.write(f'SOUR{ich}:PHAS {phase}')
    
    def get_waveform_phase(self, ich):
        self.check_channel_index(ich)
        return float(self.query(f'SOUR{ich}:PHAS?'))  # degrees

    def invert_waveform_phase(self, ich):
        ''' Invert the phase of the waveform of the specified channel. '''
        self.set_waveform_phase(ich, self.ANTIPHASE)
    
    def is_waveform_phase_inverted(self, ich):
        ''' Query whether the phase of the waveform of the specified channel is inverted. '''
        return self.get_waveform_phase(ich) == self.ANTIPHASE
    
    def set_square_duty_cycle(self, ich, DC):
        self.check_channel_index(ich)
        self.check_duty_cycle(DC)
        self.write(f'SOUR{ich}:FUNC:SQU:DCYC {DC}')

    def get_square_duty_cycle(self, ich):
        self.check_channel_index(ich)
        return float(self.query(f'SOUR{ich}:FUNC:SQU:DCYC?'))
    
    def set_pulse_hold(self, ich, phold):
        ''' 
        Set the pulse mode highlight item of the specified channel
        '''
        if phold not in self.PULSE_HOLDS:
            raise VisaError(
                f'{phold} not a valid pulse hold mode. Candidates are {self.PULSE_HOLDS}')
        self.write(f'SOUR{ich}:PULS:HOLD {phold}')
    
    def get_pulse_hold(self, ich):
        ''' 
        Get the pulse mode highlight item of the specified channel
        '''
        return self.query(f'SOUR{ich}:PULS:HOLD?')
    
    def set_pulse_duty_cycle(self, ich, DC):
        self.check_channel_index(ich)
        self.check_duty_cycle(DC)
        self.set_pulse_hold(ich, 'DUTY')
        self.write(f'SOUR{ich}:FUNC:PULS:DCYC {DC}')

    def get_pulse_duty_cycle(self, ich):
        self.check_channel_index(ich)
        return float(self.query(f'SOUR{ich}:FUNC:PULS:DCYC?'))

    def set_pulse_width(self, ich, width):
        self.check_channel_index(ich)
        self.set_pulse_hold(ich, 'WIDT')
        self.write(f'SOUR{ich}:FUNC:PULS:WIDT {width}')

    def get_pulse_width(self, ich):
        self.check_channel_index(ich)
        return float(self.query(f'SOUR{ich}:FUNC:PULS:WIDT?'))
    
    # --------------------- ARBITRARY WAVEFORM ---------------------

    def get_arbitrary_output_mode(self, ich):
        '''
        Get the output mode of the arbitrary waveform of the specified channel.

        :param ich: channel index
        :return: output mode (frequency or sample rate)
        '''
        # Check channel index
        self.check_channel_index(ich)
        # Query and return output mode
        return self.query(f'SOUR{ich}:FUNC:ARB:MODE?')
    
    def set_arbitrary_output_mode(self, ich, mode):
        '''
        Set the output mode of the arbitrary waveform of the specified channel.

        :param ich: channel index
        :param mode: output mode ("FREQ" for frequency or "SRATE" for sample rate)
        '''
        # Check channel index
        self.check_channel_index(ich)
        # Check that mode is valid
        if mode not in self.ARB_OUTPUT_MODES:
            raise VisaError(f'invalid output mode: {mode} (must be one of {self.ARB_OUTPUT_MODES})')
        # Set output mode on specified channel
        return self.write(f'SOUR{ich}:FUNC:ARB:MODE {mode}')
    
    def get_arbitrary_sample_rate(self, ich):
        '''
        Get the sample rate of the arbitrary waveform of the specified channel.

        :param ich: channel index
        :return: sample rate (Hz, or samples/s)
        '''
        # Check channel index
        self.check_channel_index(ich)
        # Query sample rate
        out = self.query(f'SOUR{ich}:FUNC:ARB:SRAT?')
        # Return as float
        return float(out)
    
    def set_arbitrary_sample_rate(self, ich, sr):
        '''
        Set the sample rate of the arbitrary waveform of the specified channel.

        :param ich: channel index
        :param sr: sample rate (Hz, or samples/s)
        '''
        # Check channel index
        self.check_channel_index(ich)
        # Check that sample rate is valid
        if not is_within(sr, self.ARB_SRATE_BOUNDS):
            raise VisaError(f'invalid sample rate: {sr} (must be within {self.ARB_SRATE_BOUNDS}')
        # Set sample rate on specified channel
        return self.write(f'SOUR{ich}:FUNC:ARB:SRAT {sr}')

    def get_waveform_catalog(self, ich=None):
        '''
        Query the arbitrary waveform data files currently stored in non-volatile memory.
        
        :param ich: channel index (optional)
        :return: list of stored waveform files
        '''
        # Get list of available waveforms as string
        query_str = 'DATA:CAT?'
        if ich is not None:
            self.check_channel_index(ich)
            query_str = f'SOUR{ich}:{query_str}'
        wfs = self.query(query_str)
        # Remove quotes and split string into list
        wfs = wfs.replace('"', '').split(',')
        # Remove empty strings
        wfs = [x for x in wfs if x != '']
        # Return list of waveforms
        return wfs

    def get_waveform_npoints(self, ich):
        '''
        Query the number of initial points of the waveform editing in
        volatile memory of specific channel.

        :param ich: channel index
        :return: number of initial points of the waveform editing
        '''
        # Check channel index
        self.check_channel_index(ich)
        # Query number of points
        n = self.query(f'SOUR{ich}:DATA:POIN? VOLATILE')
        # Return as integer
        return int(n)
    
    def set_waveform_npoints(self, ich, n):
        '''
        Set the number of initial points of the waveform editing in
        volatile memory of specific channel.

        :param ich: channel index
        :return: number of initial points of the waveform editing
        '''
        # Check channel index
        self.check_channel_index(ich)
        # Check that number of points is valid
        if not is_within(n, self.ARB_WF_NPTS_BOUNDS):
            raise VisaError(f'invalid number of points: {n} (must be within {self.ARB_WF_NPTS_BOUNDS})')
        # Set number of points for waveform editing
        self.write(f'SOUR{ich}:DATA:POIN VOLATILE,{n}')
        # Check that number of points was set correctly
        if self.get_waveform_npoints(ich) != n:
            raise VisaError(f'failed to set number of points to {n}')
    
    def write_binary_values(self, cmd, values, **kwargs):
        ''' Write binary values to instrument. '''
        logger.debug(f'{cmd} {values.size}')
        self.instrument_handle.write_binary_values(
            f'{self.PREFIX}{cmd}', values, **kwargs)

    def upload_arbitrary_waveform(self, ich, y, dtype='dac16', precision=2,
                                  adapt_npoints=True, activate=False):
        ''' 
        Upload an arbitrary waveform into volatile memory of a specific channel
        
        :param ich: channel index
        :param y: waveform vector
        :param dtype: data type used to upload waveform vector (default: "dac16")
            - "float": floating point values between -1 and 1
            - "dac": decimal DAC values between 0 and 16383
            - "dac16": 16-bits binary DAC values between 0 and 16383
        :param precision: number of decimal digits to use for each sample (default: 2, only used with "float" data type)
        :param adapt_npoints: whether to adapt the waveform vector to match the reference number of points of instrument waveforms
        :param activate: whether to set waveform type to arbitrary after upload (default: False)
        '''
        # Check channel index
        self.check_channel_index(ich)

        # Cast waveform vector to numpy array, and check waveform size
        y = np.asarray(y)
        
        # If specified, interpolate waveform vector to match the reference number
        # of points of instrument waveforms
        if adapt_npoints and y.size != self.ARB_WF_MAXNPTS_PER_PACKET:
            tref = np.arange(y.size)
            ttarget = np.linspace(tref[0], tref[-1], self.ARB_WF_MAXNPTS_PER_PACKET)
            y = np.interp(ttarget, tref, y)
        
        if not is_within(y.size, self.ARB_WF_NPTS_BOUNDS):
            raise VisaError(f'invalid waveform size: {y.size} (must be within {self.ARB_WF_NPTS_BOUNDS})')
        
        # Log
        logger.info(f'uploading {y.size}-points arbitrary waveform into volatile memory')

        # Normalize vector to appropriate range
        if dtype in ('dac', 'dac16'):
            lb, ub = self.ARB_WF_DAC_RANGE
            precision = 0  # no decimals specified for integer values
        elif dtype == 'float':
            lb, ub = self.ARB_WF_FLOAT_RANGE
        y = self.normalize(y, lb=lb, ub=ub)

        # Convert to integer if necessary
        if dtype in ('dac', 'dac16'):
            y = y.astype(int)

        # If binary upload
        if dtype == 'dac16':
            # Upload waveform to volatile memory, by successive packets
            nperpacket = self.ARB_WF_MAXNPTS_PER_PACKET
            yremain = y.copy()
            while yremain.size > 0:
                ypacket, yremain = yremain[:nperpacket], yremain[nperpacket:]
                suffix = 'CON' if yremain.size > 0 else 'END'
                self.write_binary_values(
                    f'SOUR{ich}:TRAC:DATA:DAC16 VOLATILE,{suffix},',
                    ypacket, datatype='H', is_big_endian=False)
        
        # If string upload
        else:
            # Transform waveform vector to string
            ystr = ','.join([f'{yy:.{precision}f}' for yy in y])

            # Upload waveform to volatile memory
            cmdprefix = f'SOUR{ich}:DATA'
            if dtype == 'dac':
                cmdprefix = f'{cmdprefix}:DAC'
            self.write(f'{cmdprefix} VOLATILE,{ystr}')

        # Check that waveform was uploaded correctly
        self.check_error()

        # If needed, set number of waveform points to uploaded waveform size
        # to avoid unwanted padding
        if y.size < self.ARB_WF_MAXNPTS_PER_PACKET:
            self.set_waveform_npoints(ich, y.size)

        # Set output mode to frequency to enable use as amplitude modulator
        self.set_arbitrary_output_mode(ich, 'FREQ')

        # If specified, set waveform type to arbitrary
        if activate:
            self.set_waveform_type(ich, 'USER')

    @property
    def ARB_WF_DAC_MAX(self):
        ''' Maximum integer value for arbitrary waveform DAC. '''
        return self.ARB_WF_DAC_RANGE[1]
    
    def get_waveform_value(self, ich, idx):
        '''
        Get value of arbitrary waveform loaded in volatile memory of specific channel at specific index.

        :param ich: channel index
        :param idx: waveform index (0-based, converted to 1-based for query)
        :return: waveform value at index, normalized to [0 - 1] range
        '''
        # Check channel index
        self.check_channel_index(ich)
        # Check that index is valid
        imax = self.get_waveform_npoints(ich) - 1
        if idx < 0 or idx >= imax:
            raise VisaError(f'invalid index: {idx} (must be between 0 and {imax})')
        # Query waveform value at specific index
        n = int(self.query(f'SOUR{ich}:DATA:VAL? VOLATILE,{idx + 1}'))
        # Return as float within [0 - 1] interval
        return float(n / self.ARB_WF_DAC_MAX)
    
    def set_waveform_value(self, ich, idx, val):
        '''
        Set value of arbitrary waveform loaded in volatile memory of specific channel at specific index.

        :param ich: channel index
        :param idx: waveform index (0-based, converted to 1-based for set command)
        :param val: waveform value at index, normalized to [0 - 1] range
        '''
        # Check channel index
        self.check_channel_index(ich)
        # Check that index is valid
        imax = self.get_waveform_npoints(ich) - 1
        if idx < 0 or idx >= imax:
            raise VisaError(f'invalid index: {idx} (must be between 0 and {imax})')
        # Check that value is valid
        if val < 0 or val > 1:
            raise VisaError(f'invalid value: {val} (must be between 0 and 1)')
        # Scale value to DAC integer range
        val = int(val * self.ARB_WF_DAC_MAX)
        # Set waveform value at specific index
        self.write(f'SOUR{ich}:DATA:VAL VOLATILE,{idx + 1},{val}')
        self.check_error()
    
    def download_arbitrary_waveform(self, ich):
        '''
        Get waveform vector loaded in volatile memory of specific channel.

        :param ich: channel index
        :return: waveform data as numpy array
        '''
        raise NotImplementedError('download_arbitrary_waveform() not implemented yet')
        # Check channel index
        self.check_channel_index(ich)

        # # Variant 1: binary transfer by packet (currently not working)
        # logger.info(f'querying waveform vector from channel {ich} volatile memory')
        # npackages = int(self.query(f'SOUR{ich}:DATA:LOAD? VOLATILE'))
        # data = []
        # for i in range(npackages):
        #     query = f'SOUR{ich}:DATA:LOAD? {i + 1}'
        #     chunk = self.instrument_handle.query_binary_values(
        #         f'{self.PREFIX}{query}', datatype='H', is_big_endian=False)
        #     data += chunk
        # data = np.asarray(data)

        # # Variant 2: query waveform data, point by point (currently not working)
        # npts = self.get_waveform_npoints(ich)
        # data = np.zeros(npts, dtype=int)
        # logger.info(f'querying {npts}-points waveform vector from channel {ich} volatile memory')
        # for idx in tqdm(range(npts)):
        #     val = self.query(f'SOUR{ich}:DATA:VAL? VOLATILE,{idx + 1}')
        #     self.check_error()
        #     data[idx] = int(val)
        
        # Return as vector normalized to [0 - 1] interval
        return data / self.ARB_WF_DAC_MAX
    
    def load_waveform_from_file(self, ich, name):
        '''
        Load an arbitrary waveform from file into volatile memory of a specific channel.

        :param ich: channel index
        :param name: waveform file name
        '''
        # Check channel index
        self.check_channel_index(ich)
        # Check that waveform file name is valid
        if name not in self.get_waveform_catalog(ich=ich):
            raise VisaError(f'"{name}" file not found in channel {ich} catalog')
        # Load waveform from file
        self.write(f'SOUR{ich}:DATA:COPY {name},VOLATILE')
        self.check_error()
    
    def check_waveform_file_name(self, name):
        ''' Check that a waveform file name is valid. '''
        mo = re.match(self.WAF_PATTERN, name)
        if mo is None:
            raise VisaError(f'invalid waveform file name: "{name}" (must match {self.WAF_PATTERN})')
    
    def save_waveform_to_memory(self, name):
        '''
        Store the current arbitrary waveform data (ARB) in the specified storage
        location in the internal non-volatile memory with the default name.

        :param name: waveform file name
        '''
        self.check_waveform_file_name(name)
        self.write(f'*SAV {name}')
    
    def get_waveform_from_memory(self, name):
        '''
        Recall the arbitrary waveform file (ARB) stored in the specified location
        in the internal non-volatile memory.

        :param name: waveform file name
        '''
        self.check_waveform_file_name(name)
        self.write(f'*RCL {name}')

    # --------------------- BURST ---------------------

    def enable_burst(self, ich):
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:BURS ON')
    
    def disable_burst(self, ich):
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:BURS OFF')

    def set_burst_mode(self, ich, mode):
        self.check_channel_index(ich)
        self.check_burst_mode(mode)
        self.write(f'SOUR{ich}:BURS:MODE {mode}')
    
    def get_burst_mode(self, ich):
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:BURS:MODE?')

    def set_burst_ncycles(self, ich, n):
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:BURS:NCYC {n}')

    def get_burst_ncycles(self, ich):
        self.check_channel_index(ich)
        return int(self.query(f'SOUR{ich}:BURS:NCYC?'))

    def set_burst_internal_period(self, ich, T):
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:BURS:INT:PER {T}')

    def get_burst_internal_period(self, ich):
        self.check_channel_index(ich)
        return float(self.query('SOUR{ich}:BURS:INT:PER?'))

    def set_burst_duration(self, ich, t):
        T = 1 / self.get_waveform_freq(ich)
        ncycles = self.check_burst_duration(t, T)
        self.set_burst_ncycles(ich, ncycles)
    
    def set_burst_phase(self, ich, phi):
        self.check_channel_index(ich)
        self.check_burst_phase(phi)
        self.write(f'SOUR{ich}:BURS:PHAS {phi}')

    def get_burst_phase(self, ich):
        self.check_channel_index(ich)
        return float(self.query(f'SOUR{ich}:BURS:PHAS?'))
    
    def set_burst_gated_polarity(self, ich, pol):
        self.check_channel_index(ich)
        self.check_polarity(pol)
        self.write(f'SOUR{ich}:BURS:GATE:POL {pol}')

    def get_burst_gated_polarity(self, ich):
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:BURS:GATE:POL?')

    def set_burst_delay(self, ich, t):
        ''' Set the burst delay of the N-cycle / infinite burst of the specified channel.'''
        self.write(f'SOUR{ich}:BURS:TDEL {t}')
    
    def get_burst_delay(self, ich):
        ''' Get the burst delay of the N-cycle / infinite burst of the specified channel.'''
        return float(self.query(f'SOUR{ich}:BURS:TDEL?'))
    
    def set_burst_idle_level(self, ich, lvl):
        ''' Set the idle level position of the burst mode for the specified channel '''
        self.check_channel_index(ich)
        if lvl not in self.BURST_IDLE_LEVELS:
            raise VisaError(
                f'{lvl} is not a valid idle level (options are {self.BURST_IDLE_LEVELS})')
        self.write(f'SOUR{ich}:BURS:IDLE {lvl}')
    
    def get_burst_idle_level(self, ich):
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:BURS:IDLE?')
    
    # --------------------- MODULATION ---------------------

    def enable_modulation(self, ich):
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:MOD ON')
    
    def disable_modulation(self, ich):
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:MOD OFF')

    def is_modulation_enabled(self, ich):
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:MOD?') == 'ON'

    def set_modulation_mode(self, ich, mode):
        self.check_channel_index(ich)
        self.check_modulation_mode(mode)
        self.write(f'SOUR{ich}:MOD:TYP {mode}')
    
    def get_modulation_mode(self, ich):
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:MOD:TYP?')

    @property
    def MOD_VOLT_AMP(self):
        return self.MOD_VOLT_RANGE[1] - self.MOD_VOLT_RANGE[0]
    
    # --------------------- AM ---------------------

    def get_am_depth(self, ich):
        ''' 
        Get amplitude modulation depth for the specified channel (in %)

        :param ich: channel index
        :return: modulation depth (in %)
        '''
        self.check_channel_index(ich)
        return float(self.query(f'SOUR{ich}:AM:DEPT?'))

    def set_am_depth(self, ich, depth):
        ''' 
        Set amplitude modulation depth for the specified channel (in %)

        :param ich: channel index
        :param depth: modulation depth (in %)
        '''
        self.check_channel_index(ich)
        if not is_within(depth, self.AM_DEPTH_RANGE):
            raise VisaError(f'invalid AM depth: {depth} (must be within {self.AM_DEPTH_RANGE})')
        self.write(f'SOUR{ich}:AM:DEPT {depth}')

    def enable_am_dssc(self, ich):
        ''' 
        Enable AM Distortion Suppression Signal Control (DSSC) for the specified channel. 

        :param ich: channel index
        '''
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:AM:DSSC ON')
    
    def disable_am_dssc(self, ich):
        '''
        Disable AM Distortion Suppression Signal Control (DSSC) for the specified channel. 

        :param ich: channel index
        '''
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:AM:DSSC OFF')
    
    def is_am_dssc_enabled(self, ich):
        ''' 
        Check whether AM Distortion Suppression Signal Control (DSSC) is enabled for the specified channel.

        :param ich: channel index
        '''
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:AM:DSSC?') == 'ON'
    
    def set_am_freq(self, ich, freq):
        ''' 
        Set amplitude modulation frequency for the specified channel.

        :param ich: channel index
        :param freq: modulation frequency (Hz)
        '''
        self.check_channel_index(ich)
        if not is_within(freq, self.AM_FREQ_RANGE):
            raise VisaError(f'invalid AM frequency: {freq} (must be within {self.AM_FREQ_RANGE})')
        self.write(f'SOUR{ich}:AM:INT:FREQ {freq}')
    
    def get_am_freq(self, ich):
        ''' 
        Get amplitude modulation frequency for the specified channel.

        :param ich: channel index
        :return: modulation frequency (Hz)
        '''
        self.check_channel_index(ich)
        return float(self.query(f'SOUR{ich}:AM:INT:FREQ?'))
    
    def set_am_waveform(self, ich, wtype):
        ''' 
        Set amplitude modulation waveform for the specified channel.

        :param ich: channel index
        :param wtype: modulation waveform
        '''
        self.check_channel_index(ich)
        self.check_waveform_type(wtype)
        self.write(f'SOUR{ich}:AM:INT:FUNC {wtype}')
    
    def get_am_waveform(self, ich):
        ''' 
        Get amplitude modulation waveform for the specified channel.

        :param ich: channel index
        :return: modulation waveform
        '''
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:AM:INT:FUNC?')
    
    def set_am_source(self, ich, source):
        ''' 
        Set amplitude modulation source for the specified channel.

        :param ich: channel index
        :param source: modulation source
        '''
        self.check_channel_index(ich)
        self.check_modulation_source(source)
        self.write(f'SOUR{ich}:AM:SOUR {source}')
    
    def get_am_source(self, ich):
        ''' 
        Get amplitude modulation source for the specified channel.

        :param ich: channel index
        :return: modulation source
        '''
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:AM:SOUR?')
    
    def enable_am(self, ich):
        ''' 
        Enable amplitude modulation for the specified channel.

        :param ich: channel index
        '''
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:AM:STAT ON')

    def disable_am(self, ich):
        ''' 
        Disable amplitude modulation for the specified channel.

        :param ich: channel index
        '''
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:AM:STAT OFF')
    
    def is_am_enabled(self, ich):
        ''' 
        Check whether amplitude modulation is enabled for the specified channel.

        :param ich: channel index
        '''
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:AM:STAT?') == 'ON'

    # --------------------- TRIGGER ---------------------

    def set_trigger_source(self, ich, source):
        self.check_channel_index(ich)
        self.check_trigger_source(source)
        self.write(f'SOUR{ich}:BURS:TRIG:SOUR {source}')

    def get_trigger_source(self, ich):
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:BURS:TRIG:SOUR?')

    def set_trigger_slope(self, ich, slope):
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:BURS:TRIG:SLOP {slope}')

    def get_trigger_slope(self, ich):
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:BURS:TRIG:SLOP?')
    
    def wait_for_external_trigger(self, ich):
        ''' Set up channel to wait for external trigger. '''
        self.check_channel_index(ich)
        logger.info(f'waiting for external trigger on channel {ich}...')
        self.set_trigger_source(ich, 'EXT')
        self.enable_output_channel(ich)
    
    def wait_for_manual_trigger(self, ich):
        ''' Set up channel to wait for manual trigger. '''
        self.check_channel_index(ich)
        logger.info(f'waiting for manual/programmatic trigger on channel {ich}...')
        self.set_trigger_source(ich, 'MAN')
        self.enable_output_channel(ich)

    # --------------------- TRIGGER OUTPUT ---------------------
    
    def enable_trigger_output(self, ich):
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:BURS:TRIG:TRIGO POS')

    def disable_trigger_output(self, ich):
        self.check_channel_index(ich)
        self.write(f'SOUR{ich}:BURS:TRIG:TRIGO OFF')
     
    def set_trigger_output_slope(self, ich, slope):
        self.check_channel_index(ich)
        self.check_slope(slope)
        self.write(f'SOUR{ich}:BURS:TRIG:TRIGO {slope}')

    def get_trigger_output_slope(self, ich):
        self.check_channel_index(ich)
        return self.query(f'SOUR{ich}:BURS:TRIGO:SLOP?')

    def trigger_channel(self, ich):
        ''' Trigger specific channel programmatically. '''
        self.check_channel_index(ich)
        logger.info(f'triggering channel {ich} programmatically')
        self.write(f'SOUR{ich}:BURS:TRIG:IMM')
        self.wait()

    def start_trigger_loop(self, ich, T=None):
        '''
        Start a trigger loop with a specific channel
        
        :param ich: channel index
        :param T: trigger loop period (s)
        '''
        if T is not None:
            self.set_burst_internal_period(ich, T)
        self.set_trigger_source(ich, 'INT')
    
    # --------------------- MULTI-LAYER PULSING ---------------------
    
    def set_trigger_pulse_train(self, ich, PRF, tburst, Vpp=None, T=None, trig_source='EXT'):
        '''
        Set a train of TTL-type trigger pulses on a specific channel
        
        :param ich: channel index
        :param PRF: pulse repetition frequency (Hz)
        :param tburst: burst duration (s)
        :param Vpp: pulse amplitude in V, defaults to TTL pulse amplitude (5V)
        :param T: burst repetition period in s, only for internal trigger source (defaults to 2)
        :param trig_source: trigger source (default: external)
        '''
        # Define default log message
        s = f'setting channel {ich} to trigger {si_format(tburst, 2)}s long TTL pulse train with {si_format(PRF, 2)}Hz internal PRF'

        # Set default pulse amplitude if not specified
        if Vpp is None:
            Vpp = TTL_PAMP / MV_TO_V
        else:
            s = f'{s}, {si_format(Vpp, 2)}Vpp'

        # Complete log message based on trigger source
        if trig_source == 'INT':
            if T is None:
                T = 2.  # s
            s = f'{s}, repeated every {si_format(T, 2)}s'
        elif trig_source == 'EXT':
            s = f'{s}, triggered externally'
        elif trig_source == 'MAN':
            s = f'{s}, triggered manually/programmatically'
        else:
            raise ValueError(f'invalid trigger source: {trig_source}')
        
        # Log process
        logger.info(s)
        
        # Apply pulse with specific frequency, amplitude and offset
        self.apply_pulse(ich, PRF, Vpp, offset=Vpp / 2.)
        # Set nominal pulse width
        self.set_pulse_width(ich, TTL_PWIDTH)
        # Set pulse idle level to "bottom"
        self.set_burst_idle_level(ich, 'BOTTOM')
        # Set channel trigger source to external (to avoid erroneous outputs upon setting)
        self.set_trigger_source(ich, 'EXT')
        # Set burst repetition period, if any
        if T is not None:
            self.set_burst_internal_period(ich, T)  # s
        # Set burst duration
        self.set_burst_duration(ich, tburst)  # s
        # Enable burst mode on channel
        self.enable_burst(ich)
        # Enable channel sync signal on rear panel connector
        self.enable_output_sync(ich)
        # Set channel trigger source
        self.set_trigger_source(ich, trig_source)
    
    def set_AM_pulse_train(self, ich, PRF, DC, tburst, tramp=0, T=None, trig_source='EXT'):
        '''
        Set an amplitude modulating pulse train on a specific channel
        
        :param ich: channel index
        :param PRF: pulse repetition frequency (Hz)
        :param DC: duty cycle (%)
        :param tburst: burst duration (s)
        :param tramp: nominal pulse ramping duration (s), defaults to 0
        :param T: burst repetition period in s, only for internal trigger source (defaults to 2)
        :param trig_source: trigger source (default: external)
        '''
        # Define default log message
        s = f'setting channel {ich} to trigger {si_format(tburst, 2)}s long amplitude-modulating pulse train with {si_format(PRF, 2)}Hz internal PRF'

        # Add ramping time to log message if specified
        if tramp > 0:
            s = f'{s} and {si_format(tramp, 2)}s ramping time'

        # Complete log message based on trigger source
        if trig_source == 'INT':
            if T is None:
                T = 2.  # s
            s = f'{s}, repeated every {si_format(T, 2)}s'
        elif trig_source == 'EXT':
            s = f'{s}, triggered externally'
        elif trig_source == 'MAN':
            s = f'{s}, triggered manually/programmatically'
        else:
            raise ValueError(f'invalid trigger source: {trig_source}')
        
        # Log process
        logger.info(s)

        # If ramping time is specified
        if tramp > 0:
            # Design smoothed waveform with appropriate number of points
            npts = self.ARB_WF_MAXNPTS_PER_PACKET
            _, y = get_DC_smoothed_pulse_envelope(npts, PRF, DC, tramp=tramp, plot='all')
            # Upload it to volatile memory of specified channel, 
            # and set waveform type to "user"
            self.upload_arbitrary_waveform(ich, y, activate=True)
        # Otherwise
        else:
            # Use standard rectangular pulse with specific duty cycle
            self.set_waveform_type(ich, 'SQU')
            self.set_square_duty_cycle(ich, DC)  # %
            self.invert_waveform_phase(ich)  # invert phase to avoid DC offset

        # Set waveform amplitude to full AM range (with extra margin) and ensure zero offset
        self.set_waveform_amp(ich, (1 + 2 * self.MOD_VOLT_MARGIN) * self.MOD_VOLT_AMP)
        self.set_waveform_offset(ich, 0)
        # Apply waveform as burst with specific repetition frequency
        self.set_waveform_freq(ich, PRF)
        # Set channel trigger source to external (to avoid erroneous outputs upon setting)
        self.set_trigger_source(ich, 'EXT')
        # Set burst repetition period, if any
        if T is not None:
            self.set_burst_internal_period(ich, T)  # s
        # Set burst duration
        self.set_burst_duration(ich, tburst)  # s
        # Enable burst mode on channel
        self.enable_burst(ich)
        # Enable channel sync signal on rear panel connector
        self.enable_output_sync(ich)
        # If waveform is phase-inverted, set sync polarity to negative
        if self.is_waveform_phase_inverted(ich):
            self.set_output_sync_polarity(ich, 'NEG')
        # Set channel trigger source
        self.set_trigger_source(ich, trig_source)

    def set_triggered_sine_burst_train(self, Fdrive, Vpp, tstim, PRF, DC, ich_trig=1, ich_carrier=2, **kwargs):
        '''
        Set a train of sine bursts on a specific channel, triggered by another channel. 
        Used for pulsed sinusoidal waveform generation.
        
        :param Fdrive: driving frequency (Hz)
        :param Vpp: waveform amplitude (Vpp)
        :param tstim: total stimulus duration (s)
        :param PRF: pulse repetition frequency (Hz)
        :param DC: duty cycle (%)
        :param ich_gate: index of the gating channel
        :param ich_carrier: index of the carrier channel
        '''
        # Check that channels indexes are different
        if ich_trig == ich_carrier:
            raise VisaError('trigger and carrier channels cannot be identical')
        
        # Disable all outputs (carrier channel first to avoid erroneous outputs)
        self.disable_output_channel(ich_carrier)
        self.disable_output_channel(ich_trig)
        
        # Set trigger channel parameters
        self.set_trigger_pulse_train(ich_trig, PRF, tstim, **kwargs)

        # Set carrier channel parameters
        tburst = DC / (100 * PRF)  # s
        logger.info(f'setting channel {ich_carrier} to output {si_format(tburst, 2)}s long, ({si_format(Fdrive, 2)}Hz, {si_format(Vpp, 3)}Vpp) sine wave triggered externally by channel {ich_trig}')
        self.apply_sine(ich_carrier, Fdrive, Vpp)
        self.set_burst_duration(ich_carrier, tburst)  # s
        self.enable_burst(ich_carrier)
        self.set_trigger_source(ich_carrier, 'EXT')

        # If carrier amplitude is > 0, enable all outputs 
        # (carrier channel last to avoid erroneous outputs)
        if Vpp > 0.:
            self.enable_output_channel(ich_trig)
            self.enable_output_channel(ich_carrier)
    
    def set_AM_sine_burst_train(self, Fdrive, Vpp, tstim, PRF, DC, tramp=0, ich_mod=1, ich_carrier=2, **kwargs):
        '''
        Set a train of sine bursts on a specific channel, amplitude-modulated by another channel.
        Used for pulsed sinusoidal waveform generation, with optional envelope smoothing.
        
        :param Fdrive: driving frequency (Hz)
        :param Vpp: waveform amplitude (Vpp)
        :param tstim: total stimulus duration (s)
        :param PRF: pulse repetition frequency (Hz)
        :param DC: duty cycle (%)
        :param tramp: nominal pulse ramping duration (s), defaults to 0
        :param ich_mod: index of the modulating channel
        :param ich_carrier: index of the carrier channel
        '''
        # Check that channels indexes are different
        if ich_mod == ich_carrier:
            raise VisaError('gating and carrier channels cannot be identical')
        
        # Disable all outputs (carrier channel first to avoid erroneous outputs)
        self.disable_output_channel(ich_carrier)
        self.disable_output_channel(ich_mod)
        
        # Set envelope modulating channel parameters
        self.set_AM_pulse_train(ich_mod, PRF, DC, tstim, tramp=tramp, **kwargs)

        # Set sinewave channel parameters
        logger.info(f'setting channel {ich_carrier} to output ({si_format(Fdrive, 2)}Hz, {si_format(Vpp, 3)}Vpp) sine wave amplitude-modulated externally by channel {ich_mod}')
        self.apply_sine(ich_carrier, Fdrive, Vpp, 0)
        self.enable_am(ich_carrier)
        self.set_am_source(ich_carrier, 'EXT')

        # If carrier amplitude is > 0, enable all outputs 
        # (carrier channel last to avoid erroneous outputs)
        if Vpp > 0.:
            self.enable_output_channel(ich_mod)
            self.enable_output_channel(ich_carrier)
    
    def set_gated_sine_burst(self, *args, tramp=0, ich_gate=1, ich_carrier=2, gate_type='trig', **kwargs):
        '''
        Wrapper method to set up a train of sine bursts on a carrier channel, gated by
        another channel with a specific gating method. Two gating methods are supported:
        - "trigger" gating -> calls `set_triggered_sine_burst` method
        - "mod" gating -> calls `set_AM_sine_burst` method
        The "trigger" gating method is the default, but does not support envelope smoothing.
        
        :param tramp: nominal pulse ramping duration (s), defaults to 0
        :param ich_gate: index of the gating channel
        :param ich_carrier: index of the carrier channel
        :param gate_type: gating type (default: "trigger")
        '''
        # Trigger gating
        if gate_type == 'trig':
            # Check that ramp time is 0
            if tramp > 0:
                raise VisaError('ramping time not supported for trigger gating')
            # Call `set_triggered_sine_burst` method
            self.set_triggered_sine_burst_train(
                *args, ich_trig=ich_gate, ich_carrier=ich_carrier, **kwargs)
        
        # Modulation gating
        elif gate_type == 'mod':
            # Call `set_AM_sine_burst` method
            self.set_AM_sine_burst_train(
                *args, tramp=tramp, ich_mod=ich_gate, ich_carrier=ich_carrier, **kwargs)
        
        # Invalid gating type
        else:
            raise VisaError(f'invalid gating type: {gate_type}')
    
    def set_looping_sine_burst(self, ich, Fdrive, Vpp=.1, ncycles=200, PRF=100., tramp=0, 
                               ich_trig=None, gate_type='trig'):
        '''
        Set an internally looping sine burst on a specific channel
        
        :param ich: channel index (1 or 2)
        :param Fdrive: driving frequency (Hz)
        :param Vpp: waveform amplitude (Vpp)
        :param ncycles: number of cycles per burst
        :param PRF: pulse repetition frequency (Hz)
        :param tramp: nominal pulse ramping duration (s), defaults to 0
        :param ich_trig (optional): triggering channel index
        '''
        # Cast number of cycles to integer
        ncycles = int(np.round(ncycles))
        if ncycles < 1:
            raise VisaError(f'invalid number of cycles: {ncycles} (must be >= 1)')

        # If trigger channel is different than signal channel
        if ich_trig is not None:
            # Compute nominal burst duration and duty cycle
            tburst = ncycles / Fdrive  # s

            # Compute modulation periodicity
            mod_T = 1 / PRF  # s

            # Determine gating type
            if gate_type not in ('trig', 'mod'):
                raise ValueError(f'invalid gating type: {gate_type}')
            if tramp > 0 and gate_type == 'trig':
                logger.warning('ramping time not supported for trigger gating, switching to modulation gating')
                gate_type = 'mod'

            # Determine gate-type-dependent parameters
            if gate_type == 'trig':
                DC = 100  # %
                tstim = tburst  # s
                internal_PRF = 1 / tburst
            else:
                DC = PRF * tburst * 100  # %
                if DC > 100:
                    raise ValueError(f'{si_format(tburst, 2)}s burst cannot be pulsed at {PRF:.2f} Hz')
                tstim = 2 / PRF  # s
                internal_PRF = PRF

            # Set up gated pulse train
            self.set_gated_sine_burst(
                Fdrive, Vpp, tstim, internal_PRF, DC, T=mod_T, 
                tramp=tramp if gate_type == 'mod' else 0, 
                ich_gate=ich_trig, ich_carrier=ich, gate_type=gate_type
            )
            
            # Trigger mode: start trigger loop on trigger channel
            if gate_type == 'trig':
                self.start_trigger_loop(ich_trig, T=1. / PRF)
            
            # Modulation mode: disable burst mode on modulation channel to start
            # infinite modulation loop
            else:
                self.disable_burst(ich_trig)
        
        # Otherwise, set up loop directly on signal channel
        else:
            # Check that ramping time is 0
            if tramp > 0:
                raise VisaError('ramping time not supported for looping sine burst without trigger channel')
            
            # Log process
            params_str = ', '.join([
                f'{si_format(Fdrive, 3)}Hz',
                f'{si_format(Vpp, 3)}Vpp', 
                f'{ncycles} cycles'
            ])
            logger.info(f'setting ({params_str}) sine wave looping at {PRF:.1f} Hz on channel {ich}')
            
            # Disable all outputs
            self.disable_output_channel(ich)
            
            # Set sine wave channel parameters
            self.apply_sine(ich, Fdrive, Vpp)
            self.set_burst_internal_period(ich, 1 / PRF)  # s
            self.set_burst_ncycles(ich, ncycles)
            self.enable_burst(ich)

            # Start trigger loop and enable output
            self.start_trigger_loop(ich)
            self.enable_output_channel(ich)



# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-04-07 17:51:29
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-11 17:31:16
# @Last Modified time: 2022-04-08 21:17:22

import struct
import numpy as np

from .constants import *
from .si_utils import *
from .visa_instrument import VisaError
from .oscilloscope import Oscilloscope
from .logger import logger


class RigolDS1054Z(Oscilloscope):

    # General parameters
    USB_ID = 'DS1ZA\d+' # USB identifier
    PREFIX = ':'  # prefix to be added to each command
    TIMEOUT_SECONDS = 20.  # long timeout to allow slow commands (e.g. auto-setup)
    CHANNELS = (1, 2, 3, 4)  # available channels
    NHDIVS = 12  # Number of horizontal divisions
    NVDIVS = 8  # Number of vertical divisions
    NO_ERROR_CODE = '0,"No error"'  # error code returned when no error
    MAX_VDIV = 10.  # Max voltage per division (V)
    VUNITS = ('VOLT', 'WATT', 'AMP', 'UNKN') # vertical units
    TBASES = [1, 2, 5]  # timebase relative values
    TFACTORS = np.logspace(-9, 1, 11)  # timebase multiplying factors
    TDIVS = np.ravel((np.tile(TBASES, (TFACTORS.size, 1)).T * TFACTORS).T)[2:]  # timebase values

    # Acquisition parameters
    COUPLING_MODES = (  # coupling modes
        'AC',  # alternating current
        'DC',  # direct current
        'GND'   # ground
    )
    ACQ_TYPES = ('NORM', 'AVER', 'PEAK', 'HRES')  # acquisition types
    NAVGS = np.power(2, np.arange(0, 11))  # number of sweeps to average from
    INTERP_TYPES = ('LIN', 'OFF')  # interpolation types

    # Trigger parameters
    TRIGGER_MODES = ('AUTO', 'NORM', 'SING')  # trigger modes
    TRIGGER_COUPLING_MODES = ('AC', 'DC', 'LFR', 'HFR')  # trigger coupling modes
    TRIG_TYPES = (  # trigger types
        'EDGE', 'PULS', 'RUNT', 'WIND', 'NEDG', 'SLOP', 'VID',
        'PATT', 'DEL', 'TIM', 'DUR', 'SHOL', 'RS232', 'IIC', 'SPI'
    )
    TRIGGER_SLOPES = ('POS', 'NEG', 'RFAL')  # trigger slopes

    # Waveform parameters
    SAMPLES_ON_DISPLAY = 1200  # number of samples available on display
    WAVEFORM_READING_MODES = ('NORM', 'MAX', 'RAW')   # waveform reading modes
    WAVEFORM_FORMATS = ('WORD', 'BYTE', 'ASC')  # waveform reading formats
    MAX_BYTE_LEN = 250000  # max number of bytes to read from internal memory at once

    # Units parameters
    UNITS_PER_PARAM = {}

    # --------------------- MISCELLANEOUS ---------------------

    def wait(self, t=None):
        ''' Wait for previous command to finish. '''
        self.write('*WAI')
    
    def get_last_error(self):
        ''' Get last error message. '''
        return self.query('SYST:ERR?')

    def lock_front_panel(self):
        ''' Lock front panel '''
        self.write('SYST:LOCK ON')

    def unlock_front_panel(self):
        ''' Unlock front panel '''
        self.write('SYST:LOCK OFF')

    def calibrate(self):
        ''' Calibrate oscilloscope '''
        self.query('CAL:STAR')

    def stop_calibrate(self):
        ''' Stop calibration '''
        self.query('CAL:QUIT')
    
    @staticmethod
    def decode_ieee_block(ieee_bytes):
        '''
        Strips headers (and trailing bytes) from a IEEE binary data block off.

        This is the block format commands like `WAV:DATA?`, `DISP:DATA?`,
        `SYST:SET?`, and `ETAB<n>:DATA?` return their data in.

        :param ieee_bytes: binary data block
        :return: stripped binary data block
        '''
        n_header_bytes = int(chr(ieee_bytes[1])) + 2
        n_data_bytes = int(ieee_bytes[2:n_header_bytes].decode('ascii'))
        return ieee_bytes[n_header_bytes:n_header_bytes + n_data_bytes]
    
    # --------------------- DISPLAY ---------------------

    def display_menu(self):
        pass

    def hide_menu(self):
        pass
    
    def show_trace(self, ich):
        ''' Enable trace display on specific channel '''
        self.check_channel_index(ich)
        self.write(f'CHAN{ich}:DISP ON')
    
    def hide_trace(self, ich):
        ''' Disable trace display on specific channel '''
        self.check_channel_index(ich)
        self.write(f'CHAN{ich}:DISP OFF')
    
    def is_trace(self, ich):
        ''' Query trace display on specific channel '''
        self.check_channel_index(ich)
        out = self.query(f'CHAN{ich}:DISP?')
        return bool(int(out))
    
    def get_screen_binary_img(self):
        '''
        Extract screen capture image in binary format
        
        :return: binary image
        '''
        # Query image
        self.write('DISP:DATA? ON,OFF,PNG')
        # Extract image buffer
        logger.info('Receiving screen capture...')
        buff = self.read_raw() #self.DISPLAY_DATA_BYTES)
        logger.info(f'read {len(buff)} bytes in .display_data')
        return self.decode_ieee_block(buff)

    # --------------------- SCALES / OFFSETS ---------------------

    def auto_setup(self):
        ''' Perform auto-setup. '''
        logger.info('running scope auto-setup...')
        self.write('AUT')

    def set_temporal_scale(self, value):
        ''' Set the temporal scale (in s/div) '''
        # If not in set, replace with closest valid number (in log-distance)
        if value not in self.TDIVS:
            value = self.TDIVS[np.abs(np.log(self.TDIVS) - np.log(value)).argmin()]
        logger.info(f'setting time scale to {si_format(value, 2)}s/div')
        self.write(f'TIM:MAIN:SCAL {value}')
    
    def get_temporal_scale(self):
        ''' Get the temporal scale (in s/div) '''
        out = self.query('TIM:MAIN:SCAL?')
        return float(out)
    
    def set_vertical_scale(self, ich, value):
        ''' Set the vertical sensitivity of the specified channel (in V/div) '''
        self.check_channel_index(ich)
        if value > self.MAX_VDIV:
            logger.warning(
                f'target vertical scale ({value} V/div) above instrument limit ({self.MAX_VDIV} V/div) -> restricting')
            value = self.MAX_VDIV
        logger.info(f'setting channel {ich} vertical scale to {si_format(value, 2)}V/div')
        self.write(f'CHAN{ich}:SCAL {value}')

    def get_vertical_scale(self, ich):
        ''' Get the vertical sensitivity of the specified channel (in V/div) '''
        self.check_channel_index(ich)
        out = self.query(f'CHAN{ich}:SCAL?')
        return float(out)

    def set_vertical_offset(self, ich, value):
        ''' Set the vertical offset of the specified channel (in V) '''
        self.check_channel_index(ich)
        self.write(f'CHAN{ich}:OFFS {value}')

    def get_vertical_offset(self, ich):
        ''' Get the vertical offset of the specified channel (in V) '''
        self.check_channel_index(ich)
        out = self.query(f'CHAN{ich}:OFFS?')
        return float(out)
    
    def get_vertical_unit(self, ich):
        ''' Get the vertical axis unit of a specific channel '''
        self.check_channel_index(ich)
        out = self.query(f'CHAN{ich}:UNIT?')
        return out

    def set_vertical_unit(self, ich, value):
        ''' Set the vertical axis unit of a specific channel '''
        self.check_channel_index(ich)
        if value not in self.VUNITS:
            raise VisaError(
                f'{value} not a valid vertical unit (candidates are {self.VUNITS})')
        self.write(f'CHAN{ich} {value}')
    
    def enable_vertical_scale_fine_adjustment(self, ich):
        ''' Enable fine adjustment of the vertical scale of the specific channel '''
        self.check_channel_index(ich)
        self.write(f'CHAN{ich}:VERN ON')
    
    def disable_vertical_scale_fine_adjustment(self, ich):
        ''' Disable fine adjustment of the vertical scale of the specific channel '''
        self.check_channel_index(ich)
        self.write(f'CHAN{ich}:VERN OFF')
    
    def is_vertical_scale_fine_adjustment_enabled(self, ich):
        ''' Query fine adjustment of the vertical scale of the specific channel '''
        self.check_channel_index(ich)
        out = self.query(f'CHAN{ich}:VERN?')
        return bool(int(out))
    
    def invert_vertical_scale(self, ich):
        ''' Invert the vertical scale of the specific channel '''
        self.check_channel_index(ich)
        self.write(f'CHAN{ich}:INV ON')
    
    def uninvert_vertical_scale(self, ich):
        ''' Uninvert the vertical scale of the specific channel '''
        self.check_channel_index(ich)
        self.write(f'CHAN{ich}:INV OFF')
    
    def is_vertical_scale_inverted(self, ich):
        ''' Query vertical scale inversion of the specific channel '''
        self.check_channel_index(ich)
        out = self.query(f'CHAN{ich}:INV?')
        return bool(int(out))
    
    def get_vertical_range(self, ich):
        ''' Get the vertical range of the specific channel '''
        self.check_channel_index(ich)
        out = self.query(f'CHAN{ich}:RANG?')
        return float(out)
    
    def set_vertical_range(self, ich, value):
        ''' Set the vertical range of the specific channel '''
        self.check_channel_index(ich)
        self.write(f'CHAN{ich}:RANG {value}')
    
    # --------------------- FILTERS ---------------------

    def enable_bandwith_filter(self, ich):
        ''' Turn on bandwidth-limiting low-pass filter for specific channel '''
        self.check_channel_index(ich)
        self.write(f'CHAN{ich}:BWL 20M')
    
    def disable_bandwith_filter(self, ich):
        ''' Turn off bandwidth-limiting low-pass filter for specific channel '''
        self.check_channel_index(ich)
        self.write(f'CHAN{ich}:BWL OFF')
    
    def is_bandwith_filter_enabled(self, ich):
        ''' Query bandwidth-limiting low-pass filter for specific channel '''
        self.check_channel_index(ich)
        out = self.query(f'CHAN{ich}:BWL?')
        return out == 'ON'
    
    # --------------------- PROBES & COUPLING ---------------------

    def get_probe_attenuation(self, ich):
        ''' Get the vertical attenuation factor of a specific channel '''
        self.check_channel_index(ich)
        out = self.query(f'CHAN{ich}:PROB?')
        return float(out)
    
    def set_probe_attenuation(self, ich, value):
        ''' Set the vertical attenuation factor of a specific channel '''
        self.check_channel_index(ich)
        self.write(f'CHAN{ich}:PROB {value}')

    def get_coupling_mode(self, ich):
        ''' Get the coupling mode of a specific channel '''
        return self.query(f'CHAN{ich}:COUP?')

    def set_coupling_mode(self, ich, value):
        ''' Set the coupling mode of a specific channel '''
        self.check_channel_index(ich)
        if value not in self.COUPLING_MODES:
            raise VisaError(
                f'invalid coupling mode: {value} (candidates are {self.COUPLING_MODES})')
        logger.info(f'setting channel {ich} coupling mode to {value}')
        self.write(f'CHAN{ich}:COUP {value}')

    # --------------------- TRIGGER ---------------------

    def get_trigger_mode(self):
        ''' Get trigger mode '''
        return self.query('TRIG:SWE?')
     
    def set_trigger_mode(self, value):
        ''' Set trigger mode '''
        value = value.upper()
        if value not in self.TRIGGER_MODES:
            raise VisaError(
                f'{value} not a valid trigger mode (candidates are {self.TRIGGER_MODES})')
        self.write(f'TRIG:SWE {value}')
    
    def get_trigger_coupling_mode(self, ich):
        ''' Get the trigger coupling mode. '''
        if self.get_trigger_source() != ich:
            self.set_trigger_source(ich)
        return self.query('TRIG:COUP?')
    
    def set_trigger_coupling_mode(self, ich, value):
        ''' Set the trigger coupling mode. '''
        if self.get_trigger_source() != ich:
            self.set_trigger_source(ich)
        if value not in self.TRIGGER_COUPLING_MODES:
            raise VisaError(
                f'{value} not a valid trigger coupling mode (candidates are {self.TRIGGER_COUPLING_MODES})')
        self.write(f'TRIG:COUP {value}')
    
    def get_trigger_type(self):
        ''' Get trigger type '''
        return self.query('TRIG:MODE?')
    
    def set_trigger_type(self, val):
        ''' Set trigger type '''
        if val not in self.TRIG_TYPES:
            raise VisaError(
                f'{val} not a valid trigger types. Candidates are {self.TRIG_TYPES}')
        self.write(f'TRIG:MODE {val}')
    
    def get_trigger_status(self):
        ''' Get trigger status '''
        return self.query('TRIG:STAT?')

    def set_trigger_holdoff(self, value):
        ''' Set trigger holdoff '''
        self.write(f'TRIG:HOLD {value}')
    
    def get_trigger_holdoff(self):
        ''' Get trigger holdoff '''
        return self.query('TRIG:HOLD?')
    
    def enable_trigger_noise_rejection(self):
        ''' Enable trigger noise rejection '''
        self.write('TRIG:NREJ ON')

    def disable_trigger_noise_rejection(self):
        ''' Disable trigger noise rejection '''
        self.write('TRIG:NREJ OFF')
    
    def is_trigger_noise_rejected(self):
        ''' Query trigger noise rejection status '''
        return bool(int(self.query('TRIG:NREJ?')))
    
    def get_trigger_position(self):
        ''' Get position in internal memory that corresponds to the waveform trigger position. '''
        out = int(self.query('TRIG:POS?'))
        if out == -2:
            raise VisaError('instrument not triggered')
        elif out == -1:
            raise VisaError('instrument triggered outside internal memory')
        return out

    def get_trigger_source(self):
        ''' Get trigger source channel index '''
        ttype = self.get_trigger_type()
        out = self.query(f'TRIG:{ttype}:SOUR?')
        return int(out[-1])
    
    def set_trigger_source(self, ich):
        ''' Set trigger source channel index '''
        self.check_channel_index(ich)
        logger.info(f'setting trigger source to channel {ich}')
        ttype = self.get_trigger_type()
        self.write(f'TRIG:{ttype}:SOUR CHAN{ich}')
    
    def get_trigger_slope(self, ich):
        ''' Get trigger slope for current trigger type '''
        if self.get_trigger_source() != ich:
            self.set_trigger_source(ich)
        ttype = self.get_trigger_type()
        return self.query(f'TRIG:{ttype}:SLOP?')
    
    def set_trigger_slope(self, ich, value):
        ''' Set trigger slope for current trigger type '''
        if self.get_trigger_source() != ich:
            self.set_trigger_source(ich)
        value = value.upper()
        if value not in self.TRIGGER_SLOPES:
            raise VisaError(
                f'{value} not a valid trigger slope (candidates are {self.TRIGGER_SLOPES})')
        ttype = self.get_trigger_type()
        logger.info(f'setting {ttype} trigger slope to {value}')
        self.write(f'TRIG:{ttype}:SLOP {value}')
    
    def get_trigger_level(self, ich):
        ''' Get trigger level (in V) '''
        if self.get_trigger_source() != ich:
            self.set_trigger_source(ich)
        ttype = self.get_trigger_type()
        out = self.query(f'TRIG:{ttype}:LEV?')
        return float(out)

    def set_trigger_level(self, ich, value):
        ''' Set trigger level (in V) '''
        if self.get_trigger_source() != ich:
            self.set_trigger_source(ich)
        ttype = self.get_trigger_type()
        self.write(f'TRIG:{ttype}:LEV {value}')
    
    def get_trigger_delay(self):
        ''' Get the trigger delay of the specified trigger source (in s) '''
        return float(self.query('TIM:MAIN:OFFS?'))

    def set_trigger_delay(self, value):
        ''' Set the trigger delay (in s) '''
        self.check_trigger_delay(value)
        logger.info(f'setting trigger time delay to {si_format(value, 2)}s')
        self.write(f'TIM:MAIN:OFFS {value}')

    def force_trigger(self):
        ''' Force the instrument to make 1 acquisition '''
        self.write('TFOR')
    
    # --------------------- CURSORS ---------------------
    # TODO: implement cursor functions
    
    # --------------------- ACQUISITION ---------------------
    
    def arm_acquisition(self):
        '''
        Enables the signal acquisition process by changing the acquisition state
        (trigger mode) from "stopped" to "single".
        '''
        self.write('RUN')
 
    def get_nsweeps_per_acquisition(self):
        ''' Get the number of samples to average from for average acquisition.'''
        out = self.query('ACQ:AVER?')
        return int(out)
    
    def set_nsweeps_per_acquisition(self, value):
        ''' Set the number of samples to average from for average acquisition.'''
        if value not in self.NAVGS:
            raise VisaError(f'Not a valid number of sweeps: {value}. Candidates are {self.NAVGS}')
        if value == 1:
            self.set_acquisition_type('NORM')
        else:
            self.set_acquisition_type('AVER')
            self.write(f'ACQ:AVER {value}')
    
    def get_sample_rate(self):
        ''' Get the acquisition sampling rate (in samples/second) '''
        out = self.query('ACQ:SRAT?')
        return int(float(out))
    
    def get_nsamples(self):
        ''' Get the number of samples in memory depth '''
        out = self.query('ACQ:MDEP?')
        if out == 'AUTO':
            logger.warning('auto memory depth enabled -> returning default value')
            return self.SAMPLES_ON_DISPLAY
        else:
            return int(out)

    def set_nsamples(self, value):
        ''' Set the number of samples in memory depth '''
        self.write(f'ACQ:MDEP {value}') 

    def get_acquisition_type(self):
        ''' Get oscilloscope acquisition type '''
        out = self.query('ACQ:TYPE?')
        return out
    
    def set_acquisition_type(self, value):
        ''' Set oscilloscope acquisition type '''
        value = value.upper()
        if value not in self.ACQ_TYPES:
            raise ValueError(
                f'invalid acquisition type: {value} (candidates are {self.ACQ_TYPES})')
        self.write(f'ACQ:TYPE {value}')
    
    def enable_peak_detector(self):
        ''' Turn on peak detector '''
        self.set_acquisition_type('PEAK')

    def disable_peak_detector(self):
        ''' Turn off peak detector '''
        self.set_acquisition_type('NORM')
    
    # --------------------- PARAMETERS ---------------------
    
    def get_parameter_value(self, ich, pkey):
        ''' Query the value of a parameter on a particular channel '''
        raise NotImplementedError    
    
    # --------------------- WAVEFORMS ---------------------

    def set_waveform_source(self, ich):
        ''' Set the waveform source to specific channel '''
        self.check_channel_index(ich)
        self.write(f'WAV:SOUR CHAN{ich}')

    def get_waveform_source(self):
        ''' Get the waveform source '''
        out = self.query('WAV:SOUR?')
        return int(out[-1])

    def set_waveform_reading_mode(self, value):
        ''' Set the waveform reading mode '''
        if value not in self.WAVEFORM_READING_MODES:
            raise VisaError(
                f'{value} not a valid waveform reading mode (candidates are {self.WAVEFORM_READING_MODES})')
        self.write(f'WAV:MODE {value}')
    
    def get_waveform_reading_mode(self):
        ''' Get the waveform reading mode '''
        out = self.query('WAV:MODE?')
        return out
    
    def set_waveform_format(self, value):
        ''' Set the waveform format '''
        if value not in self.WAVEFORM_FORMATS:
            raise VisaError(
                f'{value} not a valid waveform format (candidates are {self.WAVEFORM_FORMATS})')
        self.write(f'WAV:FORM {value}')
    
    def get_waveform_format(self):
        ''' Get the waveform format '''
        out = self.query('WAV:FORM?')
        return out
    
    def set_waveform_start(self, value):
        ''' Set the start point of waveform data reading. '''
        self.write(f'WAV:STAR {int(value)}')
    
    def get_waveform_start(self):
        ''' Query the start point of waveform data reading. '''
        out = self.query('WAV:STAR?')
        return int(out)
    
    def set_waveform_stop(self, value):
        ''' Set the stop point of waveform data reading. '''
        self.write(f'WAV:STOP {int(value)}')

    def get_waveform_stop(self):
        ''' Query the stop point of waveform data reading. '''
        out = self.query('WAV:STOP?')
        return int(out)
    
    def is_running(self):
        ''' Query if acquisition is running '''
        return self.get_trigger_status() in ('TD', 'WAIT', 'RUN', 'AUTO')

    def get_waveform_header(self):
        '''
        Extract a dictionary of values from the waveform header, converted to float
        and int as appropriate.

        :return: dictionary of waveform header values
        '''
        values = self.query('WAV:PRE?')
        values = values.split(',')
        assert len(values) == 10
        fmt, typ, pnts, cnt, xref, yorig, yref  = (int(val) for val in values[:4] + values[6:7] + values[8:10])
        xinc, xorig, yinc = (float(val) for val in values[4:6] + values[7:8])
        return {
            'fmt': fmt,
            'typ': typ,
            'pnts': pnts,
            'cnt': cnt,
            'xinc': xinc,
            'xorig': xorig,
            'xref': xref,
            'yinc': yinc,
            'yorig': yorig,
            'yref': yref
        }

    def get_raw_waveform_buffer(self):
        '''
        Query raw waveform buffer
        
        :return: raw waveform data buffer (in bytes)
        '''
        self.write('WAV:DATA?')
        return self.decode_ieee_block(self.read_raw())
    
    def _get_waveform_bytes(self, ich, mode='NORM'):
        '''
        Internal method to get the waveform data for a specific channel as bytes.

        This function distinguishes between requests for reading the waveform data
        currently being displayed on the screen or if you will be reading the internal 
        memory.
        
        - If you set mode to RAW, the scope will be stopped first and you will get the bytes
        from internal memory. (Please start it again yourself, if you need to, afterwards.)
        - If you set the mode to MAX this function will return the internal memory if the 
        scope is stopped, and the screen memory otherwise.

        In case the internal memory will be read, the data request will automatically be 
        split into chunks if it's impossible to read all bytes at once.

        :return: waveform data in bytes
        '''
        # Cast mode to uppercase
        mode = mode.upper()

        # Set waveform source, mode and format
        self.set_waveform_source(ich)
        self.set_waveform_reading_mode(mode)
        self.set_waveform_format('BYTE')
        
        # If normal acquisition mode, or "max" mode and acquisition is running
        if mode == 'NORM' or (mode == 'MAX' and self.is_running()):
            # Read the waveform data displayed on the screen
            return self._get_waveform_bytes_screen()

        # Otherwise
        else:
            # Read the waveform data in the internal memory
            return self._get_waveform_bytes_internal()

    def _get_waveform_bytes_screen(self):
        '''
        Internal method to extract waveform bytes from the scope if you desire
        to read the bytes corresponding to the screen content.

        :return: waveform data in bytes
        '''
        # Query raw waveform data
        buff = self.get_raw_waveform_buffer()

        # Make sure that the size of output buffer matches the number of points
        if len(buff) != self.get_waveform_header()['pnts']:
            raise VisaError('number of points in waveform buffer does not match header information')
        
        # Return waveform bytes
        return buff

    def _get_waveform_bytes_internal(self):
        '''
        Internal method to extract waveform bytes from the scope if you desire
        to read the bytes corresponding to the internal (deep) memory.
        '''        
        # If acquisition running, stop it 
        if self.is_running():
            self.stop_acquisition()

        # Extract number of points from waveform header
        npts = self.get_waveform_header()['pnts']

        # Initialize empty buffer and set starting position
        buff = b''
        pos = 1

        # Loop until all bytes have been read
        while len(buff) < npts:
            # Set waveform block start and stop position
            self.set_waveform_start(pos)
            self.set_waveform_stop(min(npts, pos + self.MAX_BYTE_LEN - 1))

            # Query waveform block and append to buffer
            buff += self.get_raw_waveform_buffer()

            # Increment position
            pos += self.MAX_BYTE_LEN
            logger.info(f'waveform acquisition: fetched {len(buff)}/{npts} points from internal memory')
        
        # Return waveform bytes
        return buff

    def get_waveform_data(self, ich, **kwargs):
        '''
        Get waveform data from a specific channel
        
        :param ich: channel index
        :return: scaled waveform data (numpy array)
        '''
        # Check channel index
        self.check_channel_index(ich)

        # Get waveform bytes and waveform header
        buff = self._get_waveform_bytes(ich, **kwargs)
        wp = self.get_waveform_header()

        # Convert bytes to samples array and rescale appropriately
        y = np.array(list(struct.unpack(str(len(buff)) + 'B', buff)))
        y = (y - wp['yorig'] - wp['yref']) * wp['yinc']

        # Get time vector
        t = np.arange(y.size) * wp['xinc'] + wp['xorig']  # s

        # Return time and voltage vectors
        return t, y




# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-04-07 17:51:29
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-11 17:26:57
# @Last Modified time: 2022-04-08 21:17:22

import re
import struct

from .constants import *
from .si_utils import *
from .utils import is_within
from .oscilloscope import Oscilloscope
from .visa_instrument import VisaError
from .logger import logger

class BK2555(Oscilloscope):

    # General parameters
    USB_ID = '378[A-Z]181\d+'  # USB identifier
    PREFIX = ''  # prefix to be added to each command
    TIMEOUT_SECONDS = 20.  # long timeout to allow slow commands (e.g. auto-setup)
    CHANNELS = (1, 2, 3, 4)  # available channels
    NHDIVS = 18  # Number of horizontal divisions
    NVDIVS = 8  # Number of vertical divisions
    NO_ERROR_CODE = 'CMR 0'  # error code returned when no error
    MAX_VDIV = 5.  # Max voltage per division (V)
    TBASES = [1, 2.5, 5]  # timebase relative values
    TFACTORS = np.logspace(-9, 1, 11)  # timebase multiplying factors
    TDIVS = np.ravel((np.tile(TBASES, (TFACTORS.size, 1)).T * TFACTORS).T)  # timebase values

    # Acquisition parameters
    COUPLING_MODES = (  # coupling modes
        'A1M',  # alternating current, 1 MOhm input impedance
        'A50',  # alternating current, 50 Ohm input impedance
        'D1M',  # direct current, 1 MOhm input impedance
        'D50',  # direct current, 50 Ohm input impedance
        'GND'   # ground
    )
    ACQ_TYPES = ('PEAK_DETECT','SAMPLING','AVERAGE')  # acquisition types
    NAVGS = (1, 4, 16, 32, 64, 128, 256)  # number of samples to average for average acquisition
    INTERP_TYPES = ('linear', 'sine')  # interpolation types

    # Trigger parameters
    TRIGGER_MODES = ('AUTO', 'NORM', 'SINGLE', 'STOP')  # trigger modes
    TRIGGER_COUPLING_MODES = ('AC', 'DC', 'HFREJ', 'LFREJ')  # trigger coupling modes
    TRIG_TYPES = (  # trigger types
        'EDGE',  # edge trigger
        'GLIT',  # pulse trigger
        'INTV',  # slope trigger
        'TV'  # video trigger
    )
    TRIGGER_SLOPES = ('NEG', 'POS', 'WINDOW')  # trigger slopes
    HOLD_TYPES = (  # pulse types
        'TI',  # holdoff
        'PS',  # if pulse width is smaller than the set value (in GLIT mode)
        'PL',  # if pulse width is larger than the set value (in GLIT mode)
        'PE',  # if pulse width is equal with the set value (in GLIT mode)
        'IS',  # if interval is smaller than the set value (in INTV mode)
        'IL',  # if interval is larger than the set value (in INTV mode)
        'IE'   # if interval is equal with the set value (in INTV mode)
    )

    # Cursor parameters
    CURSOR_TYPES = ('HREF', 'HDIF', 'VREF', 'VDIF', 'TREF', 'TDIF')  # cursor types
    CVALUES_TYPES = ('HREL', 'VREL')  # cursor value types

    # Filter parameters
    FILTER_TYPES = ('LP', 'HP', 'BP', 'BR')  # filter types
    FILTER_REL_LIMS = (2.5, 230)  # Filter cutoff limits (Hz * (s/div) = 1/div)

    # Unit parameters
    UNITS = ['S', 'V', '%', 'Hz', 'Sa']  # valid units for I/O communication
    UNITS_PER_PARAM = {
        'PKPK': 'V',  # peak-to-peak
        'MAX': 'V',  # maximum
        'MIN': 'V',  # minimum
        'AMPL': 'V',  # amplitude
        'TOP': 'V',  # top
        'BASE': 'V',  # base
        'CMEAN': 'V', # mean for cyclic waveform
        'MEAN': 'V',  # mean
        'RMS': 'V',  # root mean square
        'CRMS': 'V',  # RMS for cyclic waveform
        'OVSN': '%',  # negative overshoot 
        'OVSP': '%',  # positive overshoot 
        'FPRE': '%',  # (Vmin-Vbase)/ Vamp before the waveform falling transition
        'RPRE': '%',  # (Vmin-Vbase)/ Vamp before the waveform rising transition
        'PER': 'S',  # period
        'FREQ': 'Hz',  # frequency
        'PWID': 'S',  # positive width
        'NWID': 'S',  # negative width
        'RISE': 'S',  # rise time
        'FALL': 'S',  # fall time
        'WID': 'S',  # width
        'DUTY': '%',  # duty cycle
        'NDUTY': '%'  # negative duty cycle
    }

    # --------------------- MISCELLANEOUS ---------------------

    def wait(self, t=None):
        ''' Wait for previous command to finish. '''
        s = 'WAIT'
        if t is not None:
            s = f'{s} {t}'
        self.write(s)
    
    def get_last_error(self):
        ''' Query instrument for last error code. '''
        return self.query('CMR?')
    
    def get_status_byte_register(self):
        ''' Get the status byte register. '''
        # Query status byte register
        out = self.query('*STB?')
        # Extrat status byte register value (integer from 0 to 255)
        stb_int = self.process_int_mo(out, f'\*STB ({INT_REGEXP})', 'status byte')
        # Convert to binary string and extract status byte codes
        stb_seq = bin(stb_int)[2:].zfill(8)
        # Assemble outputs as dictionary
        stb_codes = ['INB', 'DIO1', 'VAB', 'DIO3', 'MAV', 'ESB', 'MSS/RQS', 'DIO7']
        stb_dict = {k: bool(int(s)) for k, s in zip(stb_codes, stb_seq)}
        # Remove unecessary codes
        stb_dict = {k: v for k, v in stb_dict.items() if not k.startswith('DIO')}
        # Return dictionary
        return stb_dict
    
    def lock_front_panel(self):
        ''' Lock front panel '''
        self.write('LOCK ON')

    def unlock_front_panel(self):
        ''' Unlock front panel '''
        self.write('LOCK OFF')

    def calibrate(self):
        ''' Calibrate oscilloscope '''
        self.query('*CAL?')

    def enable_automatic_calibration(self):
        ''' Enable automatic calibration. '''
        self.write('ACAL ON')

    def disable_automatic_calibration(self):
        ''' Disable automatic calibration. '''
        self.write('ACAL OFF')
    
    # --------------------- DISPLAY ---------------------

    def display_menu(self):
        ''' Display menu '''
        self.write('MENU ON')
    
    def hide_menu(self):
        ''' Hide menu '''
        self.write('MENU OFF')
    
    def show_trace(self, ich):
        ''' Enable trace display on specific channel '''
        self.check_channel_index(ich)
        self.write(f'C{ich}: TRA ON')
    
    def hide_trace(self, ich):
        ''' Disable trace display on specific channel '''
        self.check_channel_index(ich)
        self.write(f'C{ich}: TRA OFF')
    
    def is_trace(self, ich):
        ''' Query trace display on specific channel '''
        self.check_channel_index(ich)
        out = self.query(f'C{ich}: TRA?')
        return out.endswith('ON')
    
    def get_screen_binary_img(self):
        '''
        Extract screen capture image in binary format
        
        :return: binary image
        '''
        # Query image
        self.write('SCDP')
        # Extract image and return
        return self.read_raw()

    # --------------------- SCALES / OFFSETS ---------------------

    def auto_setup(self):
        ''' Perform auto-setup. '''
        logger.info('running oscilloscope auto-setup...')
        self.write('ASET')

    def set_temporal_scale(self, value):
        ''' Set the temporal scale (in s/div) '''
        # If not in set, replace with closest valid number (in log-distance)
        if value not in self.TDIVS:
            value = self.TDIVS[np.abs(np.log(self.TDIVS) - np.log(value)).argmin()]
        logger.info(f'setting time scale to {si_format(value, 2)}s/div')
        self.write(f'TDIV {self.si_process(value)}S')
    
    def get_temporal_scale(self):
        ''' Get the temporal scale (in s/div) '''
        out = self.query('TDIV?')
        return self.process_float_mo(out, f'TDIV ({SI_REGEXP})([A-z]+)', 'temporal scale')
    
    def set_vertical_scale(self, ich, value):
        ''' Set the vertical sensitivity of the specified channel (in V/div) '''
        self.check_channel_index(ich)
        if value > self.MAX_VDIV:
            logger.warning(
                f'target vertical scale ({value} V/div) above instrument limit ({self.MAX_VDIV} V/div) -> restricting')
            value = self.MAX_VDIV
        logger.info(f'setting channel {ich} vertical scale to {si_format(value, 2)}V/div')
        self.write(f'C{ich}: VDIV {self.si_process(value)}V')

    def get_vertical_scale(self, ich):
        ''' Get the vertical sensitivity of the specified channel (in V/div) '''
        self.check_channel_index(ich)
        out = self.query(f'C{ich}: VDIV?')
        return self.process_float_mo(
            out, f'C{ich}:VDIV ({SI_REGEXP})([A-z]+)', f'channel {ich} vertical scale')

    def set_vertical_offset(self, ich, value):
        ''' Set the vertical offset of the specified channel (in V) '''
        self.check_channel_index(ich)
        logger.info(f'setting channel {ich} vertical offset to {si_format(value, 2)}V/div')
        self.write(f'C{ich}: OFST {self.si_process(value)}V')

    def get_vertical_offset(self, ich):
        ''' Get the vertical offset of the specified channel (in V) '''
        self.check_channel_index(ich)
        out = self.query(f'C{ich}: OFST?')
        return self.process_float_mo(
            out, f'C{ich}:OFST ({SI_REGEXP})([A-z]+)', f'channel {ich} vertical offset')
    
    # --------------------- FILTERS ---------------------

    def enable_bandwith_filter(self, ich):
        ''' Turn on bandwidth-limiting low-pass filter for specific channel '''
        self.check_channel_index(ich)
        self.write(f'BWL C{ich}, ON')
    
    def disable_bandwith_filter(self, ich):
        ''' Turn off bandwidth-limiting low-pass filter for specific channel '''
        self.check_channel_index(ich)
        self.write(f'BWL C{ich}, OFF')
    
    def is_bandwith_filter_enabled(self, ich):
        ''' Query bandwidth-limiting low-pass filter for specific channel '''
        self.check_channel_index(ich)
        out = self.query('BWL?')
        bwl_rgxp = 'BWL ' + ','.join([f'C{c},(ON|OFF)' for c in self.CHANNELS])
        mo = re.match(bwl_rgxp, out)
        return mo.group(ich) == 'ON'
    
    def set_filter(self, ich, ftype, flow=None, fhigh=None):
        '''
        Set specific filter on specific channel
        
        :param ich: channel index
        :param ftype: filter type
        :param flow: filter lower limit frequency (Hz)
        :param flow: filter upper limit frequency (Hz)
        '''
        self.check_channel_index(ich)
        # Check filter type
        if ftype not in self.FILTER_TYPES:
            raise VisaError(
                f'invalid filter type: {ftype} (candidates are {self.FILTER_TYPES})')
        # Check frequency limits
        if ftype == 'LP':
            if flow is not None:
                raise VisaError('cannot specify frequency lower limit for LP filter')
            if fhigh is None:
                raise VisaError('frequency higher limit is required for LP filter')
        elif ftype == 'HP':
            if fhigh is not None:
                raise VisaError('cannot specify frequency higher limit for HP filter')
            if flow is None:
                raise VisaError('frequency lower limit is required for HP filter')
        else:
            if flow is None or fhigh is None:
                raise VisaError(
                    f'frequency lower and higher limits are required for {ftype} filter')
            if flow is not None and fhigh is not None:
                if flow >= fhigh:
                    raise VisaError(
                        f'frequency lower limit ({flow} Hz) must be smaller than higher limit ({fhigh} Hz)')
        tdiv = self.get_temporal_scale()  # s/div
        flims = np.asarray(self.FILTER_REL_LIMS) / tdiv  # Hz
        flims_str = ' - '.join([f'{si_format(f, 1)}Hz' for f in flims])
        for k, f in {'flow': flow, 'fhigh': fhigh}.items():
            if f is not None:
                if not is_within (f, flims):
                    raise VisaError(
                        f'{k} value ({si_format(f, 1)}Hz) outside of frequency limits ({flims_str}) with current temporal scale ({si_format(tdiv, 1)}s/div)')
        fdict = {'flow': flow, 'fhigh': fhigh}
        fdict = {k: v for k, v in fdict.items() if v is not None}
        fstr = ', '.join([f'{k} = {si_format(f, 1)}Hz' for k, f in fdict.items()])
        logger.info(f'setting {ftype} filter on channel {ich} with {fstr}')
        # Generate instruction code
        s = f'C{ich}:FILTS TYPE,{ftype}'
        if flow is not None:
            sf = si_format(flow, 1, space='').upper()
            s = f'{s},LOWLIMIT,{sf}Hz'
        if fhigh is not None:
            sf = si_format(fhigh, 1, space='').upper()
            s = f'{s},UPPLIMIT,{sf}Hz'
        self.write(s)
    
    def enable_filter(self, ich):
        ''' Turn on filter on spefific channel trace '''
        self.check_channel_index(ich)
        self.write(f'C{ich}: FILT ON')
    
    def disable_filter(self, ich):
        ''' Turn off filter on spefific channel trace '''
        self.check_channel_index(ich)
        self.write(f'C{ich}: FILT OFF')
    
    def is_filter_enabled(self, ich):
        ''' Check whether filter is enabled on spefific channel trace '''
        self.check_channel_index(ich)
        out = self.query(f'C{ich}: FILT?')
        mo = re.match(f'C{ich}:FILT (ON|OFF)', out)
        return {'ON': True, 'OFF': False}[mo.group(1)]
    
    # --------------------- PROBES & COUPLING ---------------------

    def get_probe_attenuation(self, ich):
        ''' Get the vertical attenuation factor of a specific channel '''
        self.check_channel_index(ich)
        out = self.query(f'C{ich}:ATTN?')
        mo = re.match(f'C{ich}:ATTN ({INT_REGEXP})', out)
        return float(mo[1])
    
    def set_probe_attenuation(self, ich, value):
        ''' Set the vertical attenuation factor of a specific channel '''
        self.check_channel_index(ich)
        logger.info(f'setting channel {ich} probe attenuation factor to {value}')
        self.write(f'C{ich}:ATTN {value}')

    def get_coupling_mode(self, ich):
        ''' Get the coupling mode of a specific channel '''
        out = self.query(f'C{ich}: CPL?')
        mo = re.match(f'C{ich}:CPL ([A-z0-9]+)', out)
        return mo[1]

    def set_coupling_mode(self, ich, value):
        ''' Set the coupling mode of a specific channel '''
        self.check_channel_index(ich)
        if value not in self.COUPLING_MODES:
            raise VisaError(
                f'invalid coupling mode: {value} (candidates are {self.COUPLING_MODES})')
        logger.info(f'setting channel {ich} coupling mode to {value}')
        self.write(f'C{ich}: CPL {value}')

    # --------------------- CURSORS ---------------------

    def set_cursor_position(self, ich, ctype, cpos):
        ''' 
        Set cursor position at a given screen location
        
        :param ich: cursor reference channel
        :param ctype: cursor type
        :param cpos: cursor position (in number of divisions)
        '''
        # Check cursor type
        if ctype not in self.CURSOR_TYPES:
            raise VisaError(
                f'invalid cursor type: {ctype} (candidates are {self.CURSOR_TYPES}')
        # Check cursor position
        if ctype in ('HREF', 'HDIF'):
            divbounds = (0.1, self.NHDIVS - 0.1)
        elif ctype in ('TREF', 'TDIF'):
            divbounds = (-self.NVDIVS / 2, self.NVDIVS / 2)
        elif ctype in ('VREF', 'VDIF'):
            divbounds = (-self.NHDIVS / 2, self.NHDIVS / 2)
        if not is_within(cpos, divbounds):
            raise VisaError(
                f'invalid {ctype} cursor position: {cpos} (bounds are {divbounds})')
        
        self.write(f'C{ich}: CRST {ctype}, {cpos}DIV')
    
    def get_cursor_position(self, ich, ctype):
        ''' 
        Get cursor position at a given screen location
        
        :param ich: cursor reference channel
        :param ctype: cursor type
        :return: cursor position (in number of divisions)
        '''
        # Check cursor type
        if ctype not in self.CURSOR_TYPES:
            raise VisaError(
                f'invalid cursor type: {ctype} (candidates are {self.CURSOR_TYPES}')
        out = self.query(f'C{ich}: CRST? {ctype}')
        mo = re.match(f'C{ich}:CRST {ctype},({FLOAT_REGEXP})', out)
        return float(mo.group(1))

    def get_cursor_value(self, ich, ctype):
        ''' 
        Get cursor value
        
        :param ich: cursor reference channel
        :param ctype: cursor type
        :return: cursor value(s)
        '''
        # Check cursor type
        if ctype not in self.CVALUES_TYPES:
            raise VisaError(
                f'invalid cursor type for value extraction: {ctype} (candidates are {self.CVALUES_TYPES}')
        out = self.query(f'C{ich}: CRVA? {ctype}')
        # Parser output depending on cursor type
        out_rgxp = f'C{ich}:CRVA {ctype}'
        if ctype == 'HREL':
            out_rgxp = f'{out_rgxp},({SI_REGEXP}),({SI_REGEXP}),({SI_REGEXP}),({FLOAT_REGEXP})'
        else:
            out_rgxp = f'{out_rgxp},({SI_REGEXP})'
        mo = re.match(out_rgxp, out)
        # Return cursor value(s)
        outs = [float(x) for x in mo.groups()]
        return outs

    def set_auto_cursor(self):
        ''' Set cursor mode to auto '''
        self.write('CRAU')

    # --------------------- TRIGGER ---------------------

    def get_trigger_mode(self):
        ''' Get trigger mode '''
        out = self.query('TRMD?')
        return re.match('TRMD ([A-z]+)', out)[1]
     
    def set_trigger_mode(self, value):
        ''' Set trigger mode '''
        value = value.upper()
        if value not in self.TRIGGER_MODES:
            raise VisaError(
                f'{value} not a valid trigger mode (candidates are {self.TRIGGER_MODES})')
        self.write(f'TRMD {value}')
    
    def get_trigger_coupling_mode(self, ich):
        ''' Get the trigger coupling of the selected source. '''
        self.check_channel_index(ich)
        out = self.query(f'C{ich}: TRCP?')
        return re.match(f'C{ich}:TRCP ([A-z]+)', out)[1]
    
    def set_trigger_coupling_mode(self, ich, value):
        ''' Set the trigger coupling of the selected source. '''
        self.check_channel_index(ich)
        if value not in self.TRIGGER_COUPLING_MODES:
            raise VisaError(
                f'{value} not a valid trigger coupling mode (candidates are {self.TRIGGER_COUPLING_MODES})')
        self.write(f'C{ich}: TRCP {value}')

    def get_trigger_type(self):
        ''' Get trigger type '''
        return self.get_trigger_options()['type']
    
    def set_trigger_type(self, val):
        ''' Set trigger type '''
        if val not in self.TRIG_TYPES:
            raise VisaError(
                f'{val} not a valid trigger types. Candidates are {self.TRIG_TYPES}')
        self.write(f'TRSE {val}')
    
    def get_trigger_source(self):
        ''' Get trigger source channel index '''
        return self.get_trigger_options()['source']
    
    def set_trigger_source(self, ich):
        ''' Set trigger source channel index '''
        self.check_channel_index(ich)
        logger.info(f'setting trigger source to channel {ich}')
        ttype = self.get_trigger_type()
        self.write(f'TRSE {ttype},SR,C{ich}')
    
    def get_trigger_slope(self, ich):
        ''' Get trigger slope of a particular trigger source '''
        self.check_channel_index(ich)
        out = self.query(f'C{ich}: TRSL?')
        return re.match(f'C{ich}:TRSL ([A-z]+)', out)[1]
    
    def set_trigger_slope(self, ich, value):
        ''' Set trigger slope of a particular trigger source '''
        self.check_channel_index(ich)
        value = value.upper()
        if value not in self.TRIGGER_SLOPES:
            raise VisaError(
                f'{value} not a valid trigger slope (candidates are {self.TRIGGER_SLOPES})')
        logger.info(f'setting channel {ich} trigger slope to {value}')
        self.write(f'C{ich}: TRSL {value}')

    def get_trigger_level(self, ich):
        ''' Get the trigger level of the specified trigger source (in V) '''
        self.check_channel_index(ich)
        out = self.query(f'C{ich}:TRLV?')
        return self.process_float_mo(
            out, f'C{ich}:TRLV ({SI_REGEXP})([A-z]+)', f'channel {ich} trigger level')

    def set_trigger_level(self, ich, value):
        ''' Set the trigger level of the specified trigger source (in V) '''
        self.check_channel_index(ich)
        logger.info(f'setting channel {ich} trigger level to {si_format(value, 2)}V')
        self.write(f'C{ich}:TRLV {self.si_process(value)}V')

    def set_trigger_halfamp(self):
        ''' 
        Set the trigger level of the specified trigger source to the
        centre of the signal amplitude.
        '''
        self.write('SET50')
    
    def get_trigger_delay(self):
        ''' Get trigger delay (in s) '''
        out = self.query('TRDL?')
        return self.process_float_mo(out, f'TRDL ({FLOAT_REGEXP})([A-z]+)', 'trigger delay')

    def set_trigger_delay(self, value):
        ''' Set trigger delay (in s) '''
        value = self.check_trigger_delay(value)
        logger.info(f'setting trigger time delay to {si_format(value, 2)}s')
        self.write(f'TRDL {self.si_process(value)}S')

    def get_trigger_options(self):
        ''' Get trigger type and options '''
        out = self.query(f'TRIG_SELECT?')
        tt_opts = '|'.join(self.TRIG_TYPES)
        ht_opts = '|'.join(self.HOLD_TYPES)
        mo = re.match(
            f'TRSE ({tt_opts}),SR,C({INT_REGEXP}),HT,({ht_opts}),HV,({FLOAT_REGEXP})([A-z]+)', out)
        return {
            'type': mo[1],
            'source': int(mo[2]),
            'hold_type': mo[3],
            'hold_val': self.process_float(float(mo[4]), mo[5])
        }

    def force_trigger(self):
        ''' Force the instrument to make 1 acquisition '''
        self.write('FRTR')
    
    # --------------------- ACQUISITION ---------------------
    
    def arm_acquisition(self):
        '''
        Enables the signal acquisition process by changing the acquisition state
        (trigger mode) from "stopped" to "single".
        '''
        self.write('ARM')

    def get_interpolation_type(self):
        ''' Get the type of waveform interpolation (linear or sine) '''
        out = self.query('SXSA?')
        mo = re.match('(SXSA) (ON|OFF)', out)
        _, SXSA = mo.groups()
        return {
            'ON': 'sine',
            'OFF': 'linear'
        }[SXSA]
    
    def set_interpolation_type(self, value):
        ''' Set the type of waveform interpolation (linear or sine) '''
        if value not in self.INTERP_TYPES:
            raise ValueError(
                f'invalid interpolation type: {value} (candidates are {self.INTERP_TYPES})')
        SXSA = {
            'sine': 'ON',
            'linear': 'OFF'
        }[value]
        self.write(f'SXSA {SXSA}')
    
    def get_nsweeps_per_acquisition(self):
        ''' Get the number of samples to average from for average acquisition.'''
        out = self.query('AVGA?')
        return self.process_int_mo(out, f'AVGA ({INT_REGEXP})', '#sweeps/acquisition')
    
    def set_nsweeps_per_acquisition(self, value):
        ''' Set the number of samples to average from for average acquisition.'''
        if value not in self.NAVGS:
            raise VisaError(f'Not a valid number of sweeps. Candidates are {self.NAVGS}')
        if value == 1:
            self.set_acquisition_type('SAMPLING')
        else:
            self.set_acquisition_type('AVERAGE')
            self.write(f'AVGA {value}')
    
    def get_acquisition_status(self):
        ''' Get the acquisition status of the oscilloscope '''
        out = self.query('SAST?')
        mo = re.match('SAST (.+)', out)
        return mo[1]
    
    def get_sample_rate(self):
        ''' Get the acquisition sampling rate (in samples/second) '''
        out = self.query('SARA?')
        return self.process_float_mo(out, f'SARA ({FLOAT_REGEXP})([A-z]+)', 'sample rate')
    
    def get_nsamples(self, ich):
        ''' Get the number of samples in last acquisition in a specific channel '''
        self.check_channel_index(ich)
        out = self.query(f'SANU? C{ich}')
        mo = re.match(f'SANU ({INT_REGEXP})', out)
        return int(mo[1])

    def enable_peak_detector(self):
        ''' Turn on peak detector '''
        self.write('PDET ON')

    def disable_peak_detector(self):
        ''' Turn off peak detector '''
        self.write('PDET OFF')

    def get_acquisition_type(self):
        ''' Get oscilloscope acquisition type '''
        out = self.query('ACQW?')
        mo = re.match('(ACQW) (AVERAGE,?\d*|SAMPLING|PEAK_DETECT)', out)
        _, acqtype = mo.groups()
        return acqtype
    
    def set_acquisition_type(self, value):
        ''' Set oscilloscope acquisition type '''
        value = value.upper()
        if value not in self.ACQ_TYPES:
            raise ValueError(
                f'invalid acquisition type: {value} (candidates are {self.ACQ_TYPES})')
        if value == 'AVERAGE':
            nsweeps = self.get_nsweeps_per_acquisition()
            value = f'{value},{nsweeps}'
        self.write(f'ACQW {value}')
    
    # --------------------- PARAMETERS ---------------------

    def get_parameter_value(self, ich, pkey):
        ''' Query the value of a parameter on a particular channel '''
        self.check_channel_index(ich)
        out = self.query(f'C{ich}: PAVA? {pkey}')
        if '****' in out:
            raise VisaError(f'could not extract {pkey} from channel {ich}')
        if out.endswith('%'):
            return self.process_float_mo(
                out, f'C{ich}:PAVA {pkey},({FLOAT_REGEXP})(%)', f'channel {ich} {pkey}')
        else:
            return self.process_float_mo(
                out, f'C{ich}:PAVA {pkey},({SI_REGEXP})([A-z]+)', f'channel {ich} {pkey}')
    
    def get_frequency(self, ich):
        ''' Get the waveform frequency on a particular channel '''
        self.check_channel_index(ich)
        return self.get_parameter_value(ich, 'FREQ')
    
    def get_ampl(self, ich):
        ''' Get the waveform amplitude on a particular channel '''
        self.check_channel_index(ich)
        return self.get_parameter_value(ich, 'AMPL')

    def get_min(self, ich):
        ''' Get the waveform min voltage on a particular channel '''
        self.check_channel_index(ich)
        return self.get_parameter_value(ich, 'MIN')
    
    def get_max(self, ich):
        ''' Get the waveform max voltage on a particular channel '''
        self.check_channel_index(ich)
        return self.get_parameter_value(ich, 'MAX')        

    def get_peak_to_peak(self, ich):
        ''' Get the waveform peak-to-peak voltage on a particular channel '''
        self.check_channel_index(ich)
        return self.get_parameter_value(ich, 'PKPK')
    
    # --------------------- WAVEFORMS ---------------------

    def get_waveform_settings(self):
        '''
        Get waveform settings
        
        :return: 3-tuple with (sparsing, number of points, and position of the 1st point)
        '''
        out = self.query('WAVEFORM_SETUP?')
        mo = re.match('WFSU SP,([0-9]+),NP,([0-9]+),FP,([0-9]+),SN,([0-9]+)', out)
        sp, npoints, fp, si = [int(x) for x in mo.groups()]
        return sp, npoints, fp, si
    
    @property
    def waveform_settings(self):
        sp, npoints, fp, si = self.get_waveform_settings()
        pdict = {'sparsing': sp, '# points': npoints, '1st point': fp, 'segment index': si}
        l = ['waveform settings:'] + [f' - {k} = {v}' for k, v in pdict.items()]
        return '\n'.join(l)
    
    def set_waveform_settings(self, sp=1, npoints=100, fp=0):
        '''
        Set waveform settings
        
        :param sp: sparsing,
        :param npoints: number of points
        :param fp: position of the 1st point)
        '''
        self.write(f'WFSU SP,{sp}, NP,{npoints}, FP,{fp}')
    
    def get_waveform_template(self):
        ''' Get a template description of the various logical entities making up a complete waveform.'''
        out = self.query('TMPL?')
        return out[8:-5]
    
    def get_comunication_format(self):
        '''
        Get the format features the oscilloscope uses to send waveform data
        
        :return: 3-tuple with (block_format, data_type, encoding)
        '''
        out = self.query('COMM_FORMAT?')
        mo = re.match('^CFMT (DEF9|IND0|OFF),(BYTE|WORD),(BIN|HEX)$', out)
        bfmt = mo[1]
        dtype = mo[2]
        enc = mo[3]
        return bfmt, dtype, enc

    @property
    def comunication_format(self):
        ''' Get a single string summarizing the format features the oscilloscope uses to send waveform data '''
        bfmt, dtype, enc = self.get_comunication_format()
        pdict = {'block format': bfmt, 'data type': dtype, 'encoding': enc}
        l = ['communication format:'] + [f' - {k} = {v}' for k, v in pdict.items()]
        return '\n'.join(l)

    @staticmethod
    def extract_from_bytes(bytes, istart, dtype='f'):
        '''
        Extract a field from the binary waveform header
        
        :param bytes: list of bytes representing the waveform header
        :param istart: start position of the field of interest
        :param dtype: data type of the field of interest
        :return: value of the field of interest
        '''
        # Deduce field size (in bytes) from data type
        s = {'f': 4, 'd': 8, 'l': 4}[dtype]
        
        # Extract binary segment
        fbin = b''.join(bytes[istart:istart + s])
        
        # Convert and return
        return struct.unpack(dtype, fbin)[0]

    def get_waveform_data(self, ich):
        '''
        Get waveform data from a specific channel
        
        :param ich: channel index
        :return: 2-tuple of numpy arrays with:
            - t: time signal (s)
            - y: waveform signal (V)
        '''
        # Check channel index
        self.check_channel_index(ich)
        
        # Retrieve meta data
        meta = self.query_binary_values(f'C{ich}:WF? DESC', datatype='c')
        
        # Extract number of sweeps per acquisition
        nsweeps_per_acq = self.extract_from_bytes(meta, istart=148, dtype='l')
        logger.debug(f'# sweeps/acq: {nsweeps_per_acq}')
        
        # Extract amplitude scale factor and amplitude offset
        vgain = self.extract_from_bytes(meta, istart=156)
        logger.debug(f'vertical gain: {vgain:.5e}')
        voff = self.extract_from_bytes(meta, istart=160)
        logger.debug(f'vertical offset: {voff:.5f} V')
        
        # Get sampling interval and horizontal offset
        dt = self.extract_from_bytes(meta, istart=176)
        logger.debug(f'sampling interval: {si_format(dt, 3)}s')
        hoff = self.extract_from_bytes(meta, istart=180, dtype='d')
        logger.debug(f'horizontal offset: {hoff} s')
        
        # Extract waveform data
        y = self.query_binary_values(
            f'C{ich}:WF? DAT2',
            container=np.ndarray,
            datatype='b')
        
        # Compare data length to expected number of points
        expected_npoints = self.get_waveform_settings()[1]
        if y.size != expected_npoints:
            raise VisaError(
                f'waveform parsing error: waveform size ({y.size}) does not correspond to expected number of points ({expected_npoints})')
        
        # Rescale waveform
        y = y * vgain - voff  # V
        
        # Get time vector
        t = np.arange(y.size) * dt + hoff  # s
        
        # Return
        return t, y



# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Email: theo.lemaire@epfl.ch
# @Date:   2021-02-18 18:05:57
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-22 09:52:23

from .visa_instrument import VisaError
from .waveform_generator import *
from .logger import logger


class Keysight33500B(WaveformGenerator):
    ''' High-level interface to Keysight/Agilent 33500B function generator. '''

    USB_ID = 'MY5\d+'  # USB identifier
    NO_ERROR_CODE = '+0,"No error"'  # code returned when no error
    ANGLE_UNITS = ('DEG', 'RAD')  # angle units
    MAX_LEN_TEXT = 40  # maximum length of text strings
    WAVEFORM_TYPES = ('SIN', 'SQU', 'RAMP', 'PULS', 'NOIS', 'DC', 'USER')  # waveform types
    FMAX = 10e6  # maximum frequency (Hz)
    VMAX = 10.0  # maximum voltage (V)
    BURST_MODES = ('TRIG', 'GAT')  # burst modes
    MAX_BURST_PERIOD = 10.0  # maximum burst period (s)
    TRIGGER_SOURCES = ('IMM', 'EXT', 'BUS')  # trigger sources
    CHANNELS = (1, 2)  # channels

    SWEEP_SCALES = ('LIN', 'LOG')
    MOD_FUNCS = ('SIN', 'SQU', 'RAMP', 'NRAM', 'TRI', 'NOIS', 'USER')
    MOD_SOURCES = ('INT', 'EXT')

    def beep(self):
        ''' Issue a single beep immediately. '''
        self.write('SYST:BEEP')

    # --------------------- UNITS ---------------------
    
    def set_angle_unit(self, unit):
        self.check_angle_unit(unit)
        return self.write(f'UNIT:ANGL {unit}')

    def get_angle_unit(self):
        return self.query('UNIT:ANGL?')
    
    def set_voltage_unit(self, unit):
        self.check_voltage_unit(unit)
        self.write(f'VOLT:UNIT {unit}')

    def get_voltage_unit(self):
        return self.query('VOLT:UNIT?')

    # --------------------- OUTPUT ---------------------

    def enable_output(self):
        self.write('OUTP ON')

    def disable_output(self):
        self.write('OUTP OFF')

    def get_output_state(self):
        return self.query('OUTP?')

    # --------------------- BASIC WAVEFORM ---------------------

    def apply_waveform(self, wtype, freq, amp, offset=0.):
        ''' Apply a waveform with specific parameters.

            :param wtype: waveform function
            :param freq: frequency (Hz)
            :param amp: voltage amplitude (default: Vpp)
            :param offset: voltage offset (V)
        '''
        self.check_waveform_type(wtype)
        self.write(f'APPL:{wtype} {freq}, {amp}, {offset}')
    
    def set_waveform_type(self, wtype):
        self.check_waveform_type(wtype)
        self.write(f'FUNC {wtype}')

    def get_waveform_type(self):
        return self.query('FUNC?')

    def set_waveform_freq(self, freq):
        self.check_freq(freq)
        self.write(f'FREQ {freq}')

    def get_waveform_freq(self):
        return float(self.query('FREQ?'))

    def set_waveform_amp(self, amp):
        self.check_amp(amp)
        self.write(f'VOLT {amp}')

    def get_waveform_amp(self):
        return float(self.query(f'VOLT?'))

    def set_waveform_offset(self, offset):
        self.check_offset(offset)
        self.write(f'VOLT:OFFS {offset}')

    def get_waveform_offset(self):
        return float(self.query(f'VOLT:OFFS?'))
    
    def set_square_duty_cycle(self, DC):
        self.check_duty_cycle(DC)
        self.write(f'FUNC:SQU:DCYC {DC}')

    def get_square_duty_cycle(self):
        return float(self.query('FUNC:SQU:DCYC?'))

    # --------------------- BURST ---------------------

    def enable_burst(self):
        self.write(f'BUR:STAT ON')
    
    def disable_burst(self):
        self.write(f'BUR:STAT OFF')

    def set_burst_mode(self, mode):
        self.check_burst_mode(mode)
        self.write(f'BURS:MODE {mode}')
    
    def get_burst_mode(self):
        return self.query(f'BURS:MODE?')

    def set_burst_ncycles(self, n):
        self.write(f'BURS:NCYC {n}')

    def get_burst_ncycles(self):
        return int(self.query(f'BURS:NCYC?'))

    def set_burst_internal_period(self, T):
        self.write(f'BURS:INT:PER {T}')

    def get_burst_internal_period(self):
        return float(self.query('BURS:INT:PER?'))

    def set_burst_duration(self, t):
        T = 1 / self.get_carrier_freq()
        ncycles = self.check_burst_duration(t, T)
        self.set_burst_ncycles(ncycles)

    def set_burst_phase(self, phi):
        self.check_burst_phase(phi)
        self.write(f'BURS:PHAS {phi}')

    def get_burst_phase(self):
        return float(self.query('BURS:PHAS?'))
    
    def set_burst_gated_polarity(self, pol):
        self.check_polarity(pol)
        self.write(f'BURS:GATE:POL {pol}')

    def get_burst_gated_polarity(self):
        return self.query(f'BURS:GATE:POL?')

    # --------------------- TRIGGER ---------------------

    def set_trigger_source(self, source):
        self.check_trigger_source(source)
        self.write(f'TRIG:SOUR {source}')

    def get_trigger_source(self):
        return self.query('TRIG:SOUR?')

    def set_trigger_slope(self, slope):
        self.check_slope(slope)
        self.write(f'TRIG:SLOP {slope}')

    def get_trigger_slope(self):
        return self.query('TRIG:SLOP?')
    
    def set_external_trigger(self):
        self.set_trigger_source('EXT')
        self.set_trigger_slope('POS')

    def set_internal_trigger(self):
        self.set_trigger_source('BUS')
        self.set_trigger_slope('POS')

    def single_pulse(self):
        logger.info('sending single pulse...')
        self.set_trigger_source('BUS')
        self.enable_output()
        self.trigger()
        self.wait()
    
    def wait_for_external_trigger(self):
        logger.info('waiting for external trigger...')
        self.set_trigger_source('EXT')
        self.enable_output()
    
    def wait_for_manual_trigger(self):
        logger.info('waiting for external trigger...')
        self.set_trigger_source('MAN')
        self.enable_output()

    # --------------------- TRIGGER OUTPUT ---------------------

    def enable_trigger_output(self):
        self.write('OUTP:TRIG ON')

    def disable_trigger_output(self):
        self.write('OUTP:TRIG OFF')
    
    def set_trigger_output_slope(self, slope):
        self.check_slope(slope)
        self.write(f'OUTP:TRIG:SLOP {slope}')

    def get_trigger_output_slope(self):
        return self.query('OUTP:TRIG:SLOP?')

    # --------------------- SINE ---------------------

    def set_sine_pulse(self, freq, amp, phi, t):
        ''' Set a sinusoidal pulse with specific parameters.

            :param freq: carrier frequency (Hz)
            :param amp: pulse amplitude (Vpp)
            :param phi: sine pulse phase (degrees)
            :param t: pulse duration (s)
        '''
        self.set_mod_on('BURS')
        self.set_burst_internal_period(self.MAX_BURST_PERIOD)
        self.set_carrier_freq(freq)
        self.set_carrier_amp(amp)
        if not self.get_phase_unit() == 'DEG':
            self.set_phase_unit('DEG')
        self.set_burst_phase(phi)
        self.set_burst_duration(t)

    # --------------------- PULSE ---------------------

    def set_pulse_mode(self):
        ''' Select pulse wave mode. '''
        self.set_waveform_type('PULS')

    def set_pulse_period(self, T):
        ''' Set the pulse period (s). '''
        self.write(f'PULS:PER {T}')

    def get_pulse_period(self):
        ''' Get the pulse period (s). '''
        return float(self.query('PULS:PER?'))

    # --------------------- MODULATION ---------------------

    def set_mod_on(self, key):
        ''' Enable specific modulation mode. '''
        if self.mod_key is not None and self.mod_key != key:
            self.set_mod_off(self.mod_key)
        self.write(f'{key}:STAT ON')
        self.mod_key = key

    def set_mod_off(self, key):
        ''' Disable specific modulation mode. '''
        self.write(f'{key}:STAT OFF')
        self.mod_key = None

    def get_mod_type(self):
        ''' Check which (if any) modulation mode is enabled. '''
        assert bool(self.query(f'{self.mod_key}:STAT?')), 'modulation key mismatch'
        return self.mod_key

    def set_mod_source(self, source):
        ''' Set the source type (internal or external) of a specific modulation mode. '''
        if source not in self.MOD_SOURCES:
            raise VisaError(
                f'{source} is not a valid modulating source (options are {self.MOD_SOURCES})')
        self.write(f'{self.mod_key}:SOUR {source}')

    def get_mod_source(self):
        ''' Get the source type (internal or external) of a specific modulation mode. '''
        return self.query(f'{self.mod_key}:SOUR?')

    def set_mod_func(self, func):
        ''' Set the nature of the modulating function. '''
        if func not in self.MOD_FUNCS:
            raise VisaError(
                f'{func} is not a valid modulating function (options are {self.MOD_FUNCS})')
        self.write(f'{self.mod_key}:INT:FUNC {func}')

    def get_mod_func(self):
        ''' Get the nature of the modulating function. '''
        return self.query(f'{self.mod_key}:INT:FUNC?')

    def set_mod_freq(self, freq):
        ''' Set the modulating function frequency (in Hz). '''
        self.write(f'{self.mod_key}:INT:FREQ {freq}')

    def get_mod_freq(self):
        ''' Get the modulating function frequency (in Hz). '''
        return float(self.query(f'{self.mod_key}:INT:FREQ?'))

    # --------------------- AM ---------------------

    def set_AM_mod_depth(self, depth):
        ''' Set the signal amplitude modulation depth. '''
        self.write(f'AM:DEPT {depth}')

    def get_AM_mod_depth(self):
        ''' Get the signal amplitude modulation depth. '''
        return self.query(f'AM:DEPT?')



# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-03-07 17:11:50
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-08-08 15:13:27

# Units conversion
MHZ_TO_HZ = 1e6
KHZ_TO_HZ = 1e3
PA_TO_MPA = 1e-6
S_TO_MS = 1e3
MS_TO_S = 1e-3
MV_TO_V = 1e-3
KBS_TO_BS = 1e3
S_TO_MS = 1e3
CELSIUS_TO_KELVIN = 273.15

# Default TTL pulse parameters
TTL_PWIDTH = 1e-4  # width of a nominal TTL pulse (s)
TTL_PAMP = 5.  # amplitude of a nominal TTL pulse (V)

# Regular expressions
SI_REGEXP = '[+-]?\d\.\d+?[Ee][+-]?\d+'
FLOAT_REGEXP = '[+-]?\d*[.]?\d+'
INT_REGEXP = '\d+'


# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-03-15 15:44:20
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-08-08 11:51:23

''' Initiate test sequence with Rigol waveform generator. '''

import logging
import time
import argparse

from instrulink import logger, grab_generator, VisaError

logger.setLevel(logging.INFO)

# Default waveform parameters
Fdrive = 1e4  # carrier frequency (Hz)
Vpp = 0.5  # signal amplitude (Vpp)
tstim = 200e-3  # burst duration (s)
PRF = 100.  # burst internal PRF (Hz)
DC = 50.  # burst internal duty cycle (%)
tramp = 0.  # nominal pulse ramp up time (s)

# Parse command line arguments
parser = argparse.ArgumentParser()
parser.add_argument(
    '--icarrier', type=int, default=2, choices=(1, 2), help='Carrier channel index')
parser.add_argument(
    '--igating', type=int, default=1, choices=(1, 2), help='Gating channel index')
parser.add_argument(
    '--gtype', default='trig', choices=('mod', 'trig'), 
    help='Gating type, i.e. "mod" for modulation or "trig" for trigger (only if modulation period is specified)')
parser.add_argument(
    '-s', '--source', type=str, default='int', choices=('int', 'man'), help='Trigger source (only if modulation period is specified)')
parser.add_argument(
    '-f', '--frequency', type=float, default=Fdrive, help='Carrier frequency (Hz)')
parser.add_argument(
    '--vpp', type=float, default=Vpp, help='Signal amplitude (Vpp)')
parser.add_argument(
    '--tstim', type=float, default=tstim, help='Burst duration (s)')
parser.add_argument(
    '--PRF', type=float, default=PRF, help='Burst internal PRF (Hz)')
parser.add_argument(
    '--DC', type=float, default=DC, help='Burst internal duty cycle (%)')
parser.add_argument(
    '--tramp', type=float, default=tramp, help='Nominal pulse ramp up time (ms)')
parser.add_argument(
    '-T', type=float, default=-1., help='Modulation period (s, defaults to -1, i.e. continous looping)')
args = parser.parse_args()
ich_carrier = args.icarrier  # carrier channel
ich_gate = args.igating  # gating channel
gtype = args.gtype  # gating type
trigger_source = args.source.upper()  # trigger source
Fdrive = args.frequency  # carrier frequency (Hz)
Vpp = args.vpp  # signal amplitude (Vpp)
tstim = args.tstim  # burst duration (s)
PRF = args.PRF  # burst internal PRF (Hz)
DC = args.DC  # burst internal duty cycle (%)
tramp = args.tramp * 1e-3  # nominal pulse ramp up time (s)
mod_T = args.T  # modulation period (s)

try:
    # If modulation period is not specified
    if mod_T < 0:
        # Set modulation period to None
        mod_T = None
        # If manual trigger, check that signal and gating channels are different
        if trigger_source == 'MAN' and ich_carrier == ich_gate:
            raise ValueError('Signal and gating channels must be different for single burst trigger')
    # Otherwise
    else:
        # Check that signal and gating channels are different
        if ich_carrier == ich_gate:
            raise ValueError('Signal and gating channels must be different when modulation period is specified')
except ValueError as e:
    logger.error(e)
    quit()

try:
    # Grab function generator
    wg = grab_generator()

    # If modulation period is not specified
    if mod_T is None:
        # If internal trigger is used, apply continous looping
        if trigger_source == 'INT':
            tburst = .01 * DC / PRF   # burst duration = pulse duration (s)
            wg.set_looping_sine_burst(
                ich_carrier, Fdrive, Vpp=Vpp, ncycles=tburst * Fdrive, 
                PRF=PRF, tramp=tramp, 
                ich_trig=None if ich_gate == ich_carrier else ich_gate,
                gate_type=gtype,
            )
        
        # If manual trigger is used, apply single burst
        else:
            wg.set_gated_sine_burst(
                Fdrive, Vpp, tstim, PRF, DC,
                ich_carrier=ich_carrier, ich_gate=ich_gate,
                trig_source=trigger_source,
                tramp=tramp,
                gate_type=gtype,
            )
            time.sleep(2)
            wg.trigger_channel(ich_gate)
    
    # If modulation period is specified
    else:
        # Apply signal on signal channel triggered/modulated by other channel
        wg.set_gated_sine_burst(
            Fdrive, Vpp, tstim, PRF, DC,
            tramp=tramp,
            ich_carrier=ich_carrier, ich_gate=ich_gate,
            trig_source=trigger_source, T=mod_T,
            gate_type=gtype
        )
        
        # Manual trigger: set up infinite trigger loop at specified periodicity
        if trigger_source == 'MAN':
            while True:
                wg.trigger_channel(ich_gate)
                time.sleep(mod_T)

    # Unlock front panel
    wg.unlock_front_panel()

    # Close connection
    wg.disconnect()

except VisaError as e:
    logger.error(e)
    quit()